/**
 * Backend
 */

/**
 * Конструктор карты для Яндекса
 * Настройки для карты
 * @param map_settings = {
 * field_name: $field_name (имя поля в админке)
 * service: 'yandex' (сервис провайдер)
 * map_options: {
 *      type: 'roads' (тип карты в сервис провайдере)
 *      zoom: 9 (зум по умолчанию для карты)
 *      markers: [ (массив маркеров для карты/либо 1 маркер)
 *          0 => {
 *              latitude: "55.75396",
 *              longitude: "37.620393",
 *              zoom: 9
 *          },
 *          1 => ...
 *          2 => ...
 *          ...
 *      ]
 *  }
 * }
 * @param map_object = {
 *  lat_input: $('jquery_object')
 *  lng_input: $('jquery_object')
 *  zoom_input: $('jquery_object')
 *  set_fields: function(coords, zoom)
 * }
 * @constructor
 */
function Yandex_map(map_settings, map_object) {
	if (map_object != undefined) {
		var myMap = new ymaps.Map("map-canvas-" + map_settings.field_name, {
			center: [
				map_settings.map_options.markers[0].latitude,
				map_settings.map_options.markers[0].longitude
			],
			zoom: map_settings.map_options.markers[0].zoom
		});
		var myPlacemark = this.add_markers(map_settings.map_options.markers, myMap, true);
		this.addAdminEventListeners(myMap, myPlacemark, map_object);

		return myMap;
	}
}

/**
 * Добавление маркера на карту
 *
 * @param markers
 * @param map
 * @param admin_flag
 * @returns ymaps.Placemark {}
 */
Yandex_map.prototype.add_markers = function (markers, map, admin_flag) {
	var index;

	for (index = 0; index < markers.length; index++) {
		var marker = new ymaps.Placemark(
			[
				markers[index].latitude,
				markers[index].longitude
			],
			{},
			{
				draggable: true, // метку можно перемещать
				preset: 'islands#whiteStretchyIcon'
			}
		);
		map.geoObjects.add(marker);

		if (admin_flag) {
			return marker;
		}
	}
};

/**
 * Метод ,кот. добавляет слушателей событий связанных с изменения метки/карты пользователем в админке
 *
 * @param myMap
 * @param myPlacemark
 * @param map_object
 */
Yandex_map.prototype.addAdminEventListeners = function (myMap, myPlacemark, map_object) {
	// слушатель события перемещения метки
	myPlacemark.events.add('dragend', function (e) {
		// Получение ссылки на объект, который был передвинут.
		var thisPlacemark = e.get('target');
		// Определение координат метки
		var coords = thisPlacemark.geometry.getCoordinates();
		map_object.set_fields(coords, myMap.getZoom());
	});

	// слушатель события изменения zoom'а карты
	myMap.events.add('boundschange', function (e) {
		map_object.set_fields(myPlacemark.geometry.getCoordinates(), myMap.getZoom());
	});

	// слушатель клика по карте для перемещения метки
	myMap.events.add('click', function(e){
		var target = e.get('coords');
		myPlacemark.geometry.setCoordinates(target);
		myMap.panTo(target);
		map_object.set_fields(target, myMap.getZoom());
	});

	// функция обновления местоположения маркера для событий с инпутами поля
	function inputEventsUpdate()
	{
		var coords = map_object.get_coords();
		myPlacemark.geometry.setCoordinates([coords.lat, coords.lng]);
		myMap.panTo([coords.lat, coords.lng]).then( function() {
			myMap.setZoom(coords.zoom);
		});
	}

	// обработчик изменения инпута долготы
	map_object.lat_input.on('change', function(){
		inputEventsUpdate();
	});

	// обработчик изменения инпута широты
	map_object.lng_input.on('change', function(){
		inputEventsUpdate();
	});

	// обработчик изменения инпута зума
	map_object.zoom_input.on('change', function(){
		inputEventsUpdate();
	});
};

/**
 * Конструктор карты для Гугла
 * Настройки для карты
 *
 * @param map_settings = {
 * field_name: $field_name (имя поля в админке)
 * service: 'yandex' (сервис провайдер)
 * map_options: {
 *      type: 'roads' (тип карты в сервис провайдере)
 *      zoom: 9 (зум по умолчанию для карты)
 *      markers: [ (массив маркеров для карты/либо 1 маркер)
 *          0 => {
 *              latitude: "55.75396",
 *              longitude: "37.620393",
 *              zoom: 9
 *          },
 *          1 => ...
 *          2 => ...
 *          ...
 *      ]
 *  }
 * }
 * @param map_object = {
 *  lat_input: $('jquery_object')
 *  lng_input: $('jquery_object')
 *  zoom_input: $('jquery_object')
 *  set_fields: function(coords, zoom)
 * }
 * @constructor
 */
function Google_map(map_settings, map_object) {
	if (map_object != undefined) {
		var myLatLng = new google.maps.LatLng(
			map_settings.map_options.markers[0].latitude,
			map_settings.map_options.markers[0].longitude
		);
		var map_container = document.getElementById('map-canvas-' + map_settings.field_name);
		this.map = new google.maps.Map(map_container, {
			zoom: map_settings.map_options.markers[0].zoom,
			center: myLatLng,
			mapTypeId: google.maps.MapTypeId.ROADMAP,
			timeout: 6000,
			maximumAge: 50000,
			enableHighAccuracy: true,
			zoomControl: true,
			scaleControl: true
		});
		var marker = this.add_markers(map_settings.map_options.markers, true);
		this.addAdminEventListeners(this.map, marker, map_object);

		return this;
	}
}

/**
 * Добавление маркера на карту
 *
 * @param markers
 * @param admin_flag
 * @returns google.maps.Marker {}
 */
Google_map.prototype.add_markers = function (markers, admin_flag) {
	var self = this;
	var index;

	for (index = 0; index < markers.length; index++) {
		var lat_lng = new google.maps.LatLng(markers[index].latitude, markers[index].longitude);
		var marker = new google.maps.Marker({
			position: lat_lng,
			map: self.map,
			draggable: admin_flag ? true : false,
			title: ''
		});
		marker.setMap(self.map);

		if (admin_flag) {
			return marker;
		}
	}
};

/**
 * Метод ,кот. добавляет слушателей событий связанных с изменения метки/карты пользователем в админке
 *
 * @param map
 * @param marker
 * @param map_object
 */
Google_map.prototype.addAdminEventListeners = function (map, marker, map_object) {
	// слушатель события перемещения метки по карте
	marker.addListener('dragend', function () {
		map_object.set_fields(
			[
				marker.getPosition().lat(),
				marker.getPosition().lng()
			],
			map.getZoom()
		);
	});

	// слушатель события изменения zoom'а карты
	map.addListener('zoom_changed', function () {
		map_object.set_fields(
			[
				marker.getPosition().lat(),
				marker.getPosition().lng()
			],
			this.getZoom()
		);
	});

	// слушатель клика по карте для перемещения метки
	map.addListener('click', function (e) {
		marker.setPosition(e.latLng);
		map.panTo(e.latLng);

		map_object.set_fields(
			[
				e.latLng.lat(),
				e.latLng.lng()
			],
			this.getZoom()
		);
	});

	// функция обновления местоположения маркера для событий с инпутами поля
	function inputEventsUpdate()
	{
		var coords = map_object.get_coords();
		marker.setPosition(coords);
		map.setZoom(coords.zoom);
		map.panTo(coords);
	}

	// обработчик изменения инпута долготы
	map_object.lat_input.on('change', function(){
		if($(this).val() != marker.getPosition().lat()) {
			inputEventsUpdate();
		}
	});

	// обработчик изменения инпута широты
	map_object.lng_input.on('change', function(){
		if($(this).val() != marker.getPosition().lng()) {
			inputEventsUpdate();
		}
	});

	// обработчик изменения инпута зума
	map_object.zoom_input.on('change', function(){
		if($(this).val() != map.getZoom()) {
			inputEventsUpdate();
		}
	});
};

