(function(global) {
	"use strict";
	var console = global.console || {dir: function(){}, log: function(){}, info: function() {}, warn: function() {}};

	function TAOEventManager() {
		this.registeredEvents = [];
		this.eventHandlers = [];
		this.settings = global.TAO.settings.eventManager;
	}

	TAOEventManager.prototype.registeredEvents = null;
	TAOEventManager.prototype.eventHandlers = null;
	TAOEventManager.prototype.settings = null;

	TAOEventManager.prototype.registerEvent = function (eventName) {
		var isRegistered = this.isRegistered(eventName);
		if (!isRegistered) {
			this.registeredEvents.push(eventName);
			if (this.debug) {
				console.info('TAOEventManager: Register new event [' + eventName + ']');
			}
			return this;
		}
		if (this.settings.strict) {
			throw {
				name: 'AlreadyRegisteredEventError',
				message: 'Event [' + eventName + '] is already registered'
			};
		}
		if (this.settings.debug) {
			console.warn('Event [' + eventName + '] is already registered. Can`t register second time');
		}

		return this;
	};

	TAOEventManager.prototype.isRegistered = function (eventName) {
		for (var i = 0; i < this.registeredEvents.length; i++) {
			if (eventName === this.registeredEvents[i]) {
				return true;
			}
		}
		return false;
	};

	TAOEventManager.prototype.on = function (eventName, callback, context) {
		var handler = {
			event: eventName,
			callback: callback,
			context: context
		};
		this.eventHandlers.push(handler);
		if (this.settings.debug) {
			console.info('TAOEventManager: Register new handler for event [' + eventName + ']');
			console.info(handler);
		}
		return this;
	};

	TAOEventManager.prototype.off = function (eventName, callback) {
		var foundHandler = false;
		for (var i = 0; i < this.eventHandlers.length; i++) {
			var handler = this.eventHandlers[i];
			if (handler.event === eventName && (callback === callback || !callback)) {
				this._removeHandler(i, handler);
				foundHandler = true;
			}
		}
		if (!foundHandler) {
			if (this.settings.strict) {
				throw {
					name: 'UnregisteredEventHandlerRemoveError',
					message: 'Removal of handler for event [' + eventName + '] was called, but no handler was found'
				};
			}
			if (this.settings.debug) {
				console.warn('Removal of handler for event [' + eventName + '] was called, but no handler was found');
			}
		}
		return this;
	};

	TAOEventManager.prototype._removeHandler = function (index, handler) {
		this.eventHandlers.splice(index, 1);
		if (this.settings.debug) {
			console.info('TAOEventManager: Removed handler for event [' + handler.event + ']');
			console.info(handler.callback);
		}
	};

	TAOEventManager.prototype.trigger = function (eventName, context) {
		if (this.settings.strict || this.settings.debug) {
			if (!this.isRegistered(eventName)) {
				if (this.settings.strict) {
					throw {
						name: 'UnregisteredEventTriggerError',
						message: 'Event: [' + eventName + '] was called but never registered'
					};
				}
				if (this.settings.debug) {
					console.warn('TAOEventManager: Event: [' + eventName + '] was called but never registered');
				}
			}
		}
		for (var i = 0; i < this.eventHandlers.length; i++) {
			var handler = this.eventHandlers[i];
			if (handler.event === eventName) {
				this.runHandler(handler, context, Array.prototype.slice.call(arguments, 2));
			}
		}
		if (this.settings.debug) {
			console.warn('TAOEventManager: Event: [' + eventName + '] was triggered, but no listeners found');
		}
	};

	TAOEventManager.prototype.runHandler = function (handler, context, args) {
		handler.callback.apply(handler.context ? handler.context : context, args);
		if (this.settings.debug) {
			console.info('TAOEventManager: Event: [' + handler.event + '] was triggered with context:');
			console.dir({
				context: context,
				args: args
			});
		}
	};

	global.TAO = global.TAO || {};

	global.TAO.settings = global.TAO.settings || {};

	global.TAO.settings.eventManager = global.TAO.settings.eventManager || {};

	global.TAO.events = new TAOEventManager();
})(window);