<?php

Core::load('Templates.HTML.Helpers.Maps');

/**
 * Здесь и далее используется ряд терминов требующих пояснения:
 * @текущее поле CMS.Fields - поле map_coords чей тип и valueContainer описаны в данном документе
 * @коллизия полей - при существовании двух разных полей в настройках модуля, теоретически
 * пользователь CMS может прописать этим полям идентичные свойства sqltypes, что приведет к тому, что
 * два поля будут для хранения использовать одни и те же поля в БД, что естественно приведет к потере даныых.
 * Class CMS_Fields_Types_MapCoords
 */
class CMS_Fields_Types_MapCoords extends CMS_Fields_AbstractField implements Core_ModuleInterface
{
	const VERSION = '0.0.2';
	//постфиксное имя для поля широты в базе и форме
	const LATITUDE_NAME = 'latitude';
	//постфиксное имя для поля долготы в базе и форме
	const LONGITUDE_NAME = 'longitude';
	//постфиксное имя для поля зума в базе и форме
	const ZOOM_NAME = 'zoom';
	//плейсхолдер для поля широты в шаблоне
	const LATITUDE_PLACEHOLDER = 'Широта';
	//плейсхолдер для поля долготы в шаблоне
	const LONGITUDE_PLACEHOLDER = 'Долгота';
	//плейсхолдер зума в шаблоне
	const ZOOM_PLACEHOLDER = 'Масштаб';
	//дефолтный вывод в таблице - когда не установлены координаты, либо не установлена одна из координат
	const VIEW_VALUE_DEFAULT = 'Не установлено';
	//дефолтный style
	const DEFAULT_STYLE = 'height: 500px; width: 400px; border: 1px solid #000';
	//дефолтный тип для карт
	const DEFAULT_MAP_TYPE = 'roads';

	//схема БД для создания дополнительных полей в таблице
	private static $schema = array(
		'columns' => array(
			array(
				'name' => self::LATITUDE_NAME,
				'type' => 'numeric',
				'precision' => 6,
				'scale' => 2,
				'default' => null,
			),
			array(
				'name' => self::LONGITUDE_NAME,
				'type' => 'numeric',
				'precision' => 6,
				'scale' => 2,
				'default' => null,
			),
			array(
				'name' => self::ZOOM_NAME,
				'type' => 'int',
				'default' => 9,
			),
		)
	);

	/**
	 * Метод который обрабатывает схему поля
	 * Если пользователь не указал 'sqltypes' для поля(default safe_mod предпочтительный вариант, решает проблемы коллизией
	 * полей):
	 * Формирует схему в соответствии со схемой поля map_coords, приписывая полям префикс - название поля в настройках
	 * компонента.
	 * Если пользователь ввел свои sqltypes:
	 * Стандартная сборка полей для sqltypes, правда с условием что массив sqltypes может состоять только из 1 или 3х полей:
	 * 1 - в поле будет храниться serialize строка значений lat;lng;zoom
	 * 3 - в каждом поле будет хранится своя величина lat|lng|zoom
	 *
	 * @param string $field_name
	 * @param array $data
	 * @return array
	 */
	public function schema($field_name, $data)
	{
		if (!isset($data['schema'])) {
			$data['schema'] = array();
		}

		if (!isset($data['sqltypes']) && !isset($data['sqltype'])) {
			$data['schema']['columns'][$field_name] = array(
				'name' => $field_name,
				'type' => 'text',
				'default' => null,
			);
		} else {
			if (count($data['sqltypes']) > 3) {
				$data['sqltypes'] = array_slice($data['sqltypes'], 0, 3);
			} elseif (count($data['sqltypes']) == 2) {
				$data['sqltypes'] = array_slice($data['sqltypes'], 0, 1);
			}
		}

		return $data['schema'];
	}

	/**
	 * Добавляем необходимые для карты поля к форме
	 *
	 * @param Forms_Form $form
	 * @param string $name
	 * @param array $data
	 */
	public function form_fields($form, $name, $data)
	{
		if ($langs = $this->data_langs($data)) {
			foreach ($langs as $lang => $ldata) {
				foreach (self::$schema['columns'] as $value) {
					$form->input($name . '_' . $value['name'] . '_' . $lang);
				}
			}
		} else {
			foreach (self::$schema['columns'] as $value) {
				$form->input($name . '_' . $value['name']);
			}
		}

		return parent::form_fields($form, $name, $data);
	}


	/**
	 * Настраиваем вывод значения map_coords в таблице списка записей
	 *
	 * @param CMS_ORM_Entity $value
	 * @param string $name
	 * @param array $data
	 * @param bool|false $row
	 * @return string
	 */
	public function view_value($value, $name, $data, $row = false)
	{
		if (!isset($data['sqltypes'])) {
			return !empty($value[$name]) ? $value[$name] : self::VIEW_VALUE_DEFAULT;
		} else {
			if (count($data['sqltypes']) == 1) {
				return !empty($value[key($data['sqltypes'])]) ? $value[key($data['sqltypes'])] : self::VIEW_VALUE_DEFAULT;
			} else {
				$array_keys = array_keys($data['sqltypes']);
				$view_value = '';

				foreach ($array_keys as $key) {
					if (!empty($value[$key])) {
						$view_value .= $value[$key] . ';';
					}
				}

				return !empty($view_value) ? substr($view_value, 0, strlen($view_value) - 1) : self::VIEW_VALUE_DEFAULT;
			}
		}
	}

	/**
	 * Присвоение сохраненных координат в админке при открытии формы редактирования
	 *
	 * @param $form
	 * @param $object
	 * @param $name
	 * @param $data
	 */
	public function assign_from_object($form, $object, $name, $data)
	{
		$form_fields = $this->get_form_fields($name);

		if (isset($data['sqltypes'])) {
			if (count($data['sqltypes']) == 1) {
				$values = explode(';', $object[key($data['sqltypes'])]);
				$i = 0;

				foreach ($form_fields as $value) {
					$form[$value['name']] = floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($values[$i]));
					$i++;
				}
			} else {
				$i = 0;
				$array_keys = array_keys($data['sqltypes']);

				foreach ($form_fields as $value) {
					$form[$value['name']] = floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($object[$array_keys[$i]]));
					$i++;
				}
			}
		} else {
			$values = explode(';', $object[$name]);
			$i = 0;

			foreach ($form_fields as $value) {
				$form[$value['name']] = floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($values[$i]));
				$i++;
			}
		}
	}

	/**
	 * Сохранение координат в админке
	 *
	 * @param $form
	 * @param $object
	 * @param $name
	 * @param $data
	 */
	public function assign_to_object($form, $object, $name, $data)
	{
		$form_fields = $this->get_form_fields($name);

		if (isset($data['sqltypes'])) {
			if (count($data['sqltypes']) == 1) {
				$values = array();

				foreach ($form_fields as $value) {
					$values[] = floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($form[$value['name']]));
				}

				$object[key($data['sqltypes'])] = implode(';', $values);
			} else {
				$i = 0;
				$sqltypes_keys = array_keys($data['sqltypes']);

				foreach ($form_fields as $value) {
					$object[$sqltypes_keys[$i]] = floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($form[$value['name']]));
					$i++;
				}
			}
		} else {
			$values = array();

			foreach ($form_fields as $value) {
				$values[] = floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($form[$value['name']]));
			}

			$object[$name] = implode(';', $values);

		}
	}


	/**
	 * Добавляем скрипты, стили и настройки скриптов в layout
	 *
	 * @param Templates_HTML_Template $layout
	 * @param string $name
	 * @param array $data
	 * @return mixed
	 */
	protected function layout_preprocess($layout, $name, $data)
	{
		$layout_preprocessor = new CMS_Fields_Types_MapCoords_Layout_Preprocessor($layout, $name, $data);

		$layout_preprocessor->prepare_layout();

		return parent::layout_preprocess($layout_preprocessor->get_layout(), $name, $data);
	}

	/**
	 * Добавляем опции в шаблон(требуются для настройки отображения шаблона)
	 *
	 * @param Templates_Template $template
	 * @param string $name
	 * @param array $data
	 * @return mixed
	 */
	protected function preprocess($template, $name, $data)
	{
		/**
		 * @var Templates_Template $template
		 */
		$template = parent::preprocess($template, $name, $data);

		$options = array(
			'properties' => array(
				'style' => isset($data['properties']['style']) ? $data['properties']['style'] : self::DEFAULT_STYLE,
				'enable_search' => isset($data['properties']['enable_search']) ? $data['map_options']['enable_search'] : false,
				'show_zoom' => isset($data['show_zoom']) ? (bool)$data['show_zoom'] : false,
			),
		);


		return $template->with(array('options' => $options));
	}

	/**
	 * Возвращаем стандартный массив полей, который используется в форме
	 *
	 * @param string $name - имя текущего поля в CMS.Fields
	 * @return array
	 */
	private function get_form_fields($name)
	{
		return array(
			array(
				'name' => $name . '_' . self::LATITUDE_NAME,
			),
			array(
				'name' => $name . '_' . self::LONGITUDE_NAME,
			),
			array(
				'name' => $name . '_' . self::ZOOM_NAME,
			),
		);
	}
}

class CMS_Fields_Types_MapCoords_ValueContainer extends CMS_Fields_ValueContainer
{
	//formats are deprecated since version 0.0.2
	protected $formats = array(
		'degrees_ru' => array(
			'lat' => array('с.ш', 'ю.ш'),
			'lng' => array('в.д', 'з.д'),
		),
		'degrees_en' => array(
			'lat' => array('N', 'S'),
			'lng' => array('E', 'W'),
		),
		'degrees' => array(
			'lat' => array('', ''),
			'lng' => array('', ''),
		),
	);

	/**
	 * @var CMS_Fields_Types_MapCoords
	 */
	protected $type;

	private $schema;

	/**
	 * @param string $name
	 * @param array $data
	 * @param stdClass $item
	 * @param CMS_Fields_Types_MapCoords $type
	 */
	public function __construct($name, $data, $item, $type)
	{
		parent::__construct($name, $data, $item, $type);
		$this->schema = $this->type->schema($name, $this->data);
	}

	/**
	 * Вывод значения поля map_coords с помощью valueContainer
	 *
	 * @param null|string $part
	 * @param null|string $format
	 *
	 * @return array|string
	 */
	public function value($part = null, $format = null)
	{
		if (isset($this->data['sqltypes'])) {
			if (count($this->data['sqltypes']) == 1) {
				$data = $this->get_data_from_string($this->item[key($this->data['sqltypes'])]);
				$value = $this->get_value_by_part($data, $part, $format);
			} else {
				$array_keys = array_keys($this->data['sqltypes']);
				$data = array(
					'lat' => $this->item[$array_keys[0]],
					'lng' => $this->item[$array_keys[1]],
					'zoom' => $this->item[$array_keys[2]]
				);
				$value = $this->get_value_by_part($data, $part, $format);
			}
		} else {
			$values = explode(';', $this->item[$this->name]);
			$data = array(
				'lat' => CMS_Fields_Types_MapCoords_Helper::floatize_string($values[0]),
				'lng' => CMS_Fields_Types_MapCoords_Helper::floatize_string($values[1]),
				'zoom' => $values[2]
			);
			$value = $this->get_value_by_part($data, $part, $format);
		}

		return $value;
	}

	/**
	 * format is now deprecated, лучше не использовать, т.к. функционал не поддерживаемый переехал на новую версию
	 * для сохранения совместимости
	 *
	 * @param null|string $format
	 * @param string $template
	 *
	 * @return array|string
	 */
	public function render($format = null, $template = '{lat}, {lng}')
	{
		switch ($format) {
			case 'degrees_en':
			case 'degrees_ru':
			case 'degrees':
			case 'decimal':
				$value = $this->value('latlng');
				if ($value) {
					if ($value['lat'] > 0) {
						$lat_suff = ' ' . $this->formats[$format]['lat'][0];
					} else {
						$value['lat'] = -$value['lat'];
						$lat_suff = ' ' . $this->formats[$format]['lat'][1];
					}
					if ($value['lng'] > 0) {
						$lng_suff = ' ' . $this->formats[$format]['lng'][0];
					} else {
						$value['lng'] = -$value['lng'];
						$lng_suff = ' ' . $this->formats[$format]['lng'][1];
					}

					if ($format == 'decimal') {
						$lat = (string)$value['lat'];
						$lng = (string)$value['lng'];
					} else {
						$lat_degree = $this->decimals_to_degrees($value['lat']);
						$lng_degree = $this->decimals_to_degrees($value['lng']);
						$lat = $lat_degree . $lat_suff;
						$lng = $lng_degree . $lng_suff;
					}
					$value = $template;
					$value = preg_replace('/{lat}/', $lat, $value);
					$value = preg_replace('/{lng}/', $lng, $value);
					if ($value === null) {
						$value = $lat . ', ' . $lng;
					}
				}
				break;
			default:
				$value = $this->value();
				break;
		}
		if (is_null($value)) {
			return '';
		}
		if (!is_string($value)) {
			return print_r($value, true);
		}
		return $value;
	}

	/**
	 * Превращение градусов в десятичные координаты
	 *
	 * @param string $degree
	 *
	 * @return float|int|null
	 */
	public function degrees_to_decimals($degree)
	{
		$degree = CMS_Fields_Types_MapCoords_Helper::floatize_string($degree);
		if ($degree == '') {
			$decimal = null;
		} else {
			if (is_numeric($degree)) {
				$decimal = floatval($degree);
			} else {
				$pattern = "/^(?:(-?\d+)°)*\s*(?:(\d+)')*\s*(?:(\d+)'')*$/";
				if (preg_match($pattern, trim($degree), $m)) {

					$deg = (int)$m[1];
					$min = (int)$m[2];
					$sec = (int)$m[3];

					$sign = 1;
					if ($deg < 0) {
						$deg *= -1;
						$sign = -1;
					}

					$decimal = $sign * ($deg + $min / 60 + $sec / 3600);

				} else {
					$decimal = 0;
				}
			}
		}
		return $decimal;
	}

	/**
	 * Deprecated function format
	 * Возвращает значение в необходимом формате(десятичные/градусы)
	 *
	 * @param string $value
	 * @param string $part
	 * @param string $format
	 *
	 * @return string
	 */
	protected function format($value, $part, $format)
	{
		if (array_key_exists($format, $this->formats)) {
			if ($value >= 0) {
				$value = $this->decimals_to_degrees($value);
				$value = $value . ' ' . $this->formats[$format][$part][0];
			} else {
				$value = -$value;
				$value = $this->decimals_to_degrees($value);
				$value = $value . ' ' . $this->formats[$format][$part][1];
			}
		}
		return $value;
	}

	/**
	 * Метод возвращает значение поля в соответствии с запросом части и формата значения
	 *
	 * @param array $data
	 * @param string $part
	 * @param null|string $format deprecated параметр, можно использовать, но поддержки больше нет
	 *
	 * @return array|string
	 */
	private function get_value_by_part($data, $part, $format = null)
	{
		switch ($part) {
			case 'lat':
			case 'lng':
				$value = $this->format($data[$part], $part, $format);
				break;
			case 'latlng':
				$value['lat'] = $this->format($data['lat'], 'lat', $format);
				$value['lng'] = $this->format($data['lng'], 'lng', $format);
				break;
			case 'zoom':
				$value = $data['zoom'];
				break;
			default:
				$data['lat'] = $this->format($data['lat'], 'lat', $format);
				$data['lng'] = $this->format($data['lng'], 'lng', $format);
				$value = $data['lat'] . ";" . $data['lng'] . ';' . $data['zoom'];
				break;
		}

		return $value;
	}

	/**
	 * Сделать массив данных из строки
	 * В метод передается строка вида lat;lng;zoom
	 *
	 * @param string $str
	 *
	 * @return array
	 */
	private function get_data_from_string($str)
	{
		$data = explode(';', $str);

		return array(
			'lat' => CMS_Fields_Types_MapCoords_Helper::floatize_string($data[0]),
			'lng' => CMS_Fields_Types_MapCoords_Helper::floatize_string($data[1]),
			'zoom' => $data[2]
		);
	}

	/**
	 * Превращение десятичных координат в градусы для вывода
	 *
	 * @param string $decimal
	 *
	 * @return string
	 */
	private function decimals_to_degrees($decimal)
	{
		$val = "";
		$sign = '';
		if ($decimal < 0) {
			$decimal *= -1;
			$sign = '-';
		}

		$deg = floor($decimal);
		$min_val = ($decimal - $deg) * 60;
		$min = floor($min_val);
		$sec = round(($min_val - $min) * 60);

		$val .= $sign . $deg . "° ";
		if ($min > 0) {
			$val .= $min . "' ";
		}
		if ($sec > 0) {
			$val .= $sec . "''";
		}
		return $val;
	}

	/**
	 * Механизм проверки наличие контента для рендера в контейнере
	 *
	 * @param null|string $lang
	 * @return bool
	 */
	public function check_empty($lang = null)
	{
		return !$this->value('lat') || !$this->value('lng');
	}
}

class CMS_Fields_Types_MapCoords_Layout_Preprocessor
{
	private static $admin_scripts = array('fields/map-coords.js', 'admin/maps.js', 'jquery/popup/colorbox.js');
	private static $admin_css = array('fields/map-coords.css', 'jquery/colorbox/colorbox.css');
	/**
	 * @var Templates_HTML_Template
	 */
	private $layout;
	/**
	 * @var string
	 */
	private $field_name;
	/**
	 * @var array
	 */
	private $field_data;

	/**
	 * @param Templates_HTML_Template $layout
	 * @param string $name
	 * @param array $data
	 */
	public function __construct($layout, $name, $data)
	{
		$this->layout = $layout;
		$this->field_name = $name;
		$this->field_data = $data;
	}

	/**
	 * Подготовка шаблона состоит из двух этапов:
	 * 1. подключение стандартных скриптов и css
	 * 2. генерация специфичных скриптов и данных для карт.
	 * @return void
	 */
	public function prepare_layout()
	{
		$this->default_layout_preprocess();
		$this->map_scripts_preprocess();
	}

	/**
	 * @return Templates_HTML_Template
	 */
	public function get_layout()
	{
		return $this->layout;
	}

	/**
	 * Добавляем в layout базовые стили и скрипты
	 */
	private function default_layout_preprocess()
	{
		$this->layout->use_scripts(self::$admin_scripts);
		$this->layout->use_styles(self::$admin_css);
	}

	/**
	 * Запускаем препроцессинг шаблона в котором будет отображено поле координат
	 * @return void
	 */
	private function map_scripts_preprocess()
	{
		$map_layout_preprocessor = new CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor($this->layout, $this->field_name, $this->field_data);
		$map_layout_preprocessor->set_layout_map_scripts();
	}
}

class CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor
{
	//дефолтный сервис для отображения карты
	const GOOGLE_MAP_SERVICE = 'google';
	//альтернативные карты
	const YANDEX_MAP_SERVICE = 'yandex';
	/**
	 * @var string
	 */
	private $service_name;
	/**
	 * @var CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_Helper
	 */
	private $service;
	/**
	 * @var Templates_HTML_Template
	 */
	private $layout;
	/**
	 * @var string
	 */
	private $fields_name;
	/**
	 * @var array
	 */
	private $field_data;
	/**
	 * @var CMS_Fields_Types_MapCoords_Map_Options
	 */
	private $map_options;

	/**
	 * @param Templates_HTML_Template $layout
	 * @param $fields_name
	 * @param $field_data
	 */
	public function __construct(&$layout, $fields_name, $field_data)
	{
		$this->layout = $layout;
		$this->fields_name = $fields_name;
		$this->field_data = $field_data;
		$this->service_name = $field_data['service'];
		$this->map_options = new CMS_Fields_Types_MapCoords_Map_Options($fields_name, $field_data);
		$this->set_map_layout_helper();
	}

	/**
	 * Исходя из настроек поля выбираем хелпер для подключения скриптов карты
	 *
	 * @return void
	 */
	private function set_map_layout_helper()
	{
		if ($this->service_name == self::YANDEX_MAP_SERVICE) {
			$this->service = new CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_YandexMap_Helper($this->layout);
		} else {
			$this->service = new CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_GoogleMap_Helper($this->layout);
		}
	}

	/**
	 * @return Templates_HTML_Template
	 */
	public function get_layout()
	{
		return $this->layout;
	}

	/**
	 * Подключаем стандартные скрипты карт методом из хелпера
	 * После чего генерируем данные для скриптов и пишем напрямую в шаблон
	 *
	 * @return void
	 */
	public function set_layout_map_scripts()
	{
		$this->layout = $this->service->setup_scripts()->get_layout();
		$this->map_options->set_map_options_to_template($this->service->get_service_name());
	}
}

abstract class  CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_Helper
{
	/**
	 * @var Templates_HTML_Template
	 */
	protected $layout;
	protected $service_name = '';

	/**
	 * @param Templates_HTML_Template $layout
	 */
	public function __construct($layout)
	{
		$this->layout = $layout;
	}

	/**
	 * @return Templates_HTML_Template
	 */
	public function get_layout()
	{
		return $this->layout;
	}

	/**
	 * @return string
	 */
	public function get_service_name()
	{
		return $this->service_name;
	}

	/**
	 * @return $this
	 */
	abstract public function setup_scripts();
}


class CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_YandexMap_Helper extends CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_Helper
{
	//подключение js api для yandex
	const YANDEX_JS_API_STRING = 'https://api-maps.yandex.ru/2.1/?lang=ru_RU&onload=initMapsYandex';

	protected $service_name = 'yandex';

	public function setup_scripts()
	{
		$url =  self::YANDEX_JS_API_STRING;
		if ($key = CMS::vars()->get('ymaps_key')) {
			$url .= '&apikey='. $key;
		}
		$this->layout->use_script($url);
		return $this;
	}

}

class CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_GoogleMap_Helper extends CMS_Fields_Types_MapCoords_Layout_Map_Preprocessor_Helper
{
	//подключение js api для google
	const GOOGLE_JS_API_STRING = 'https://maps.googleapis.com/maps/api/js?sensor=false&callback=initMapsGoogle';

	protected $service_name = 'google';



	public function setup_scripts()
	{
		$this->layout->use_script(
			self::GOOGLE_JS_API_STRING,
			array(
				'attrs' => array(
					'async' => 'async',
					'defer' => 'defer'
				)
			)
		);

		return $this;
	}

}

class CMS_Fields_Types_MapCoords_Map_Options
{
	//дефолтная широта(Москва)
	const DEFAULT_LATITUDE = 55.75396;
	//дефолтная долгота(Москва)
	const DEFAULT_LONGITUDE = 37.620393;
	//дефолтный зум для карты
	const DEFAULT_ZOOM = 9;

	/**
	 * @var array
	 */
	private $options;
	/**
	 * @var array
	 */
	private $field_data;
	/**
	 * @var string
	 */
	private $field_name;

	/**
	 * @param $field_name
	 * @param $field_data
	 */
	public function __construct($field_name, $field_data)
	{
		$this->options = array('field_name' => $field_name);
		$this->field_data = $field_data;
		$this->field_name = $field_name;
		$this->generate_settings();
	}

	/**
	 * Прописываем в шаблон настройки/опции для карты
	 * @param $service
	 */
	public function set_map_options_to_template($service)
	{
		Templates_HTML::add_scripts_settings(
			array(
				'map' => array(
					$this->field_name => array_merge_recursive(
						$this->options,
						array(
							'service' => $service
						)
					)
				)
			)
		);
	}

	/**
	 * Генерируем массив настроек для карты
	 * (по умолчанию если координат нет то передаются координаты Москвы:
	 * latitude: 55.75396, longitude: 37.620393, zoom: 9)
	 *
	 * @return array
	 */
	private function generate_settings()
	{
		if (!isset($this->field_data['sqltypes'])) {
			//создаем ключи для поиска значение в $this->field_data['__item']
			$values = explode(';', $this->field_data['__item'][$this->field_name]);
			//в массиве $data по ключу __item лежит ссылка на текущую запись в БД
			$settings['map_options']['markers'][0] = array(
				'latitude' => empty($values[0]) ? self::DEFAULT_LATITUDE : floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($values[0])),
				'longitude' => empty($values[1]) ? self::DEFAULT_LONGITUDE : floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($values[1])),
				'zoom' => empty($values[2]) ? self::DEFAULT_ZOOM : (int)$values[2],
			);
		} else {
			if (count($this->field_data['sqltypes']) == 1) {
				$values = explode(';', $this->field_data['__item'][key($this->field_data['sqltypes'])]);
				$settings['map_options']['markers'][0] = array(
					'latitude' => empty($values[0]) ? self::DEFAULT_LATITUDE : floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($values[0])),
					'longitude' => empty($values[1]) ? self::DEFAULT_LONGITUDE : floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($values[1])),
					'zoom' => empty($values[2]) ? self::DEFAULT_ZOOM : (int)$values[2],
				);
			} else {
				$array_keys = array_keys($this->field_data['sqltypes']);
				$settings['map_options']['markers'][0] = array(
					'latitude' => empty($this->field_data['__item'][$array_keys[0]]) ? self::DEFAULT_LATITUDE : floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($this->field_data['__item'][$array_keys[0]])),
					'longitude' => empty($this->field_data['__item'][$array_keys[1]]) ? self::DEFAULT_LONGITUDE : floatval(CMS_Fields_Types_MapCoords_Helper::floatize_string($this->field_data['__item'][$array_keys[1]])),
					'zoom' => empty($this->field_data['__item'][$array_keys[2]]) ? self::DEFAULT_ZOOM : (int)$this->field_data['__item'][$array_keys[2]],
				);
			}
		}

		$this->options = array_merge_recursive($this->options, $settings);
	}
}

class CMS_Fields_Types_MapCoords_Helper
{
	/**
	 * Преобразование строки, соджержащей float число в зависимости от локали, установленной на сервере
	 * в разных локалях, разный delimiter, т.о. можно потерять дробную часть после разделителя
	 *
	 * @param $float_string
	 * @return mixed
	 */
	public static function floatize_string($float_string)
	{
		$locale_info = localeconv();
		$thousands_sep = $locale_info['thousands_sep'] ? $locale_info['thousands_sep'] : $locale_info['mon_thousands_sep'];
		$decimal_point = $locale_info['decimal_point'] ? $locale_info['decimal_point'] : $locale_info['mon_decimal_point'];
		$float_string = str_replace($thousands_sep, "", $float_string);
		$float_string = str_replace($decimal_point, ".", $float_string);
		return $float_string;
	}
}
