<?php
/**
 * @package CMS\Fields\ValueContainer
 *
 * Class CMS_Fields_ValueContainer
 *
 * @property string                        $name
 * @property array                         $data
 * @property CMS_ORM_Entity                $item
 * @property CMS_Fields_AbstractField|null $type
 * @property CMS_Lang|string|bool          $forced_lang
 * @property bool                          $force_current_lang
 */
class CMS_Fields_ValueContainer implements Core_ModuleInterface
{
	protected $name;
	protected $data;
	protected $item;
	protected $type;

	protected $forced_lang = false;

	protected static $force_current_lang = false;

	public static function initialize($config)
	{
		if (isset($config['force_current_lang'])) {
			self::$force_current_lang = $config['force_current_lang'];
		}
	}

	public function __construct($name, $data, $item, $type)
	{
		$this->name = $name;
		$this->data = $data;
		$this->item = $item;
		$this->type = $type;

		$this->setup();
	}

	protected function setup()
	{
		if ($this->force_current_lang_is_required()) {
			$this->forced_lang = CMS::site_lang();
		}
	}

	protected function force_current_lang_is_required()
	{
		return isset($this->data['force_current_lang']) ? $this->data['force_current_lang'] : self::$force_current_lang;
	}


	protected function template()
	{
		$template = CMS::layout_view();
		if (!$template) {
			$template = Templates_HTML::Template('empty');
		}
		return $template;
	}

	public function value()
	{
		return $this->item->{$this->name};
	}

	/**
	 * @param string      $value
	 * @param null|string $lang
	 * @return CMS_Lang|string
	 */
	protected function parse_multilang_value($value, $lang = null)
	{
		if (is_string($value) && $this->type->enable_multilang()) {
			if ($lang !== null) {
				return CMS::lang($value, $lang);
			}
			if ($this->forced_lang) {
				return CMS::lang($value, $this->forced_lang);
			}
			return CMS::lang($value);
		}
		return $value;
	}

	/**
	 * @param null|string $lang
	 * @return CMS_Lang|mixed|string
	 */
	public function multilang_value($lang = null)
	{
		return $this->parse_multilang_value($this->value(), $lang);
	}

	public function set($value)
	{
		$this->item->{$this->name} = $value;
		return $this;
	}

	/**
	 * Копирует значение из указанного объекта в текущий
	 *
	 * @param CMS_ORM_Entity $from
	 *
	 * @return $this
	 */
	public function copy_value($from)
	{
		$this->type->copy_value($from, $this->item, $this->name, $this->data);
		return $this;
	}

	public function lang($lang = false)
	{
		$this->forced_lang = $lang ? $lang : CMS::site_lang();
		return $this;
	}

	public function change_data(array $data)
	{
		$this->data = array_merge($this->data, $data);
		return $this;
	}

	public function draw_in_layout($layout = false, $template = 'template', $parms = array())
	{
		return $this->type->draw_in_layout($this->name, $this->data, $layout, $template, $parms);
	}

	public function render()
	{
		if (!$this->type->access($this->name, $this->data, 'container_render', $this->item, $this)) {
			return '';
		}
		$value = $this->value();
		if (is_null($value)) {
			return '';
		}
		if (!is_string($value)) {
			return print_r($value, true);
		}
		return $this->multilang_value();
	}

	protected function value_to_path($value)
	{
		if ($value == '') {
			return false;
		}
		if ($value[0] == '.' || $value[0] == '/') {
			return $value;
		}
		return "./$value";
	}

	public function path()
	{
		return $this->value_to_path(trim($this->value()));
	}

	public function value_to_url($value)
	{
		if ($value == '') {
			return false;
		}
		$value = preg_replace('{^\./}', '', $value);
		if ($value[0] == '.' || $value[0] == '/') {
			return $this->type->controller($this->name, $this->data)->download_url($value, $this->item, $this->name, $this->data);
		}
		return "/$value";
	}

	public function url()
	{
		return $this->value_to_url(trim($this->value()));
	}

	/**
	 * Проверка на отсутствие контента для рендера в контейнере.
	 * Для переопредления механизма проверки используется check_empty
	 *
	 * @param null|string $lang
	 * @return bool
	 */
	public function is_empty($lang = null)
	{
		if (isset($this->data['is_empty']) && Core_Types::is_callable($this->data['is_empty'])) {
			return Core::invoke($this->data['is_empty'], array($this, $this->item, $lang));
		}
		return $this->check_empty($lang);
	}

	/**
	 * Механизм проверки наличие контента для рендера в контейнере
	 *
	 * @param null|string $lang
	 * @return bool
	 */
	public function check_empty($lang = null)
	{
		return !$this->multilang_value($lang);
	}

	/**
	 * @param null|string $lang
	 * @return bool
	 */
	public function is_not_empty($lang = null)
	{
		return !$this->is_empty($lang);
	}

	public function __toString()
	{
		return (string)$this->render();
	}

	public function __get($name)
	{
		if (isset($this->$name)) {
			return $this->$name;
		}
		throw new Core_MissingPropertyException($name);
	}

	public function access($action)
	{
		return $this->type->access($this->name, $this->data, $action, $this->item);
	}

	public function __call($method, $args)
	{
		if (method_exists($this->type, $method)) {
			call_user_func_array(array($this->type, $method), $args);
			return $this;
		}
		throw new Core_MissingMethodException($method);
	}
}