<?php

class CMS_Packages implements Core_ModuleInterface
{

	protected static $options = array(
		'dir' => '../app/packages',
		'auto_install' => false,
		'name' => 'CMSPackages',
		'admin_menu_title' => 'Frontend пакеты',
		'admin_menu_icon' => 'bookshelf.png',
		'admin_path' => 'packages',
	);

	protected static $instances = array();

	public static function initialize($options = array())
	{
		self::$options = array_replace_recursive(self::$options, $options);
		CMS::add_component(self::$options['name'], self::router());
		CMS_Admin::menu(self::$options['admin_menu_title'], CMS::admin_path(self::$options['admin_path']), self::$options['admin_menu_icon']);
		if (!empty(WS::env()->config->packages)) {
			self::$options = array_merge(self::$options, (array) WS::env()->config->packages);
		}
	}

	protected static function router()
	{
		return Core::make('CMS.Packages.Router');
	}

	public static function option($name)
	{
		return self::$options[$name];
	}

	public static function name_to_module($name)
	{
		$name = ucwords(strtolower($name));
		foreach (array('-', '_', '.') as $delimiter) {
			if (strpos($name, $delimiter) !== false) {
				$name = implode('', array_map('ucfirst', explode($delimiter, $name)));
			}
		}
		return $name;
	}

	public function reset_instances($name = null)
	{
		if (is_null($name)) {
			self::$instances = array();
		} else {
			unset(self::$instances[(string) $name]);
		}
	}

	public static function find($name)
	{
		if (isset(self::$instances[$name])) {
			return self::$instances[$name];
		}
		$dirs = array_filter(glob(self::option('dir') . '/*'), 'is_dir');
		foreach ($dirs as $dir) {
			$base_name = basename($dir);
			if ($name == $base_name) {
				$module = self::name_to_module($name);
				$file = $dir . '/' . $module . '.php';
				if (IO_FS::exists($file)) {
					return self::load($name, $module, $file);
				}
			}
		}
		$inst = self::make_empty($name);
		if ($inst->exists()) {
			return self::$instances[$name] = self::make_empty($name);
		}
		return null;
	}

	public static function load($name, $module, $file)
	{
		if (isset(self::$instances[$name])) {
			return self::$instances[$name];
		}

		self::include_file($file);
		$instance = self::make_instance($module, $name);
		self::$instances[$name] = $instance;
		return $instance;
	}

	protected static function include_file($file)
	{
		return include_once($file);
	}

	public static function make_instance($module, $name)
	{
		$class = "\TAO\Packages\\{$module}";
		return new $class($name);
	}

	public static function make_empty($name)
	{
		return new CMS_Packages_Instance($name);
	}

	public static function auto_install($name, $config = array())
	{
		if (!self::option('auto_install')) {
			return;
		}
		$stockroom = CMS_Stockroom::manager();
		$list = $stockroom->components_list(array('search' => $name));
		if($list) {
			foreach ($list as $entity) {
				if (
					/*Core_Strings::contains($entity->gitdir, $name) && */
					!$entity->installed()) {
					if ($stockroom->validate_package_intall($entity)) {
						$stockroom->install_package_files($entity);
						return self::find($name);
					}
				}
			}
		}
		return null;
	}

}

class CMS_Packages_Instance
{

	protected $name;
	protected $docroot_dir = './packages';
	protected $find_patterns = array(
		'styles' => array('!.*\.css$!', '!.*\.less$!', '!.*\.scss$!'),
		'scripts' => array('!.*\.js$!')
	);
	protected $files = array();

	public function __construct($name)
	{
		$this->name = $name;
	}

	public function exists()
	{
		return IO_FS::exists($this->dir());
	}

	public function dir($path = '')
	{
		$result = $this->docroot_dir . '/' . $this->name . '/' . $path;
		return rtrim($result, '/');
	}

	public function read_dir($dir)
	{
		$files = array();
		$iterator = new RecursiveIteratorIterator(new RecursiveDirectoryIterator($dir));
		foreach ($iterator as $key => $file) {
			if ($file->isDir()){ 
				continue;
			}
			$files[$key] = $file;
		}
		return $files;
	}

	public function filter_app_files($files)
	{
		$to_remove = array();
		foreach ($files as $path => $file) {
			if (Core_Strings::starts_with($path, $this->dir() . '/app')) {
				$clear_path = str_replace('/app', '', $path);
				$to_remove[] = isset($files[$clear_path]) ? $clear_path : $path;
			}
		}
		foreach ($to_remove as $key) {
			unset($files[$key]);
		}
		return $files;
	}

	public function filter_files_by_patterns($files)
	{
		$filtered = array();
		foreach ($files as $path => $file) {
			foreach ($this->find_patterns as $name => $patterns) {
				foreach ($patterns as $pattern) {
					if ($this->filter_by_pattern($path, $pattern)) {
						$filtered[$name][basename($path)] = $path;
					}
				}
			}
		}
		return $filtered;
	}

	public function filter_by_pattern($path, $pattern)
	{
		return preg_match($pattern, $path);
	}

	protected function cache_key($key)
	{
		return "tao:packages:{$this->name}:{$key}";
	}

	protected function cache_get($key)
	{
		$key = $this->cache_key($key);
		return WS::env()->cache->get($key);
	}

	protected function cache_set($key, $value, $lifetime = 3600)
	{
		$key = $this->cache_key($key);
		return WS::env()->cache->set($key, $value, $lifetime);
	}

	protected function find_files_to_use($use_cache = true)
	{
		$files = array();
		if (!$use_cache || !($files = $this->cache_get('files'))) {
			$files = $this->read_dir($this->dir());
			$files = $this->filter_app_files($files);
			$files = $this->filter_files_by_patterns($files);
			if ($use_cache) {
				$this->cache_set('files', $files);
			}
		}
		$this->files = $files;
		return $files;
	}

	public function file_by_name($name, $type = 'scripts')
	{
		return isset($this->files[$type][$name]) ? $this->files[$type][$name] : null;
	}

	protected function use_files($view, $files, $config = array())
	{
		foreach ($files as $name => $value) {
			$urls = array();
			foreach ($value as $key => $filename) {
				$urls[$key] = array_merge(array('name' => '/' . ltrim($filename, '/.')), $config);
			}
			$method = "use_{$name}";
			$view->{$method}($urls);
		}
	}

	public function use_in_template($view, $config = array())
	{
		$files = $this->find_files_to_use();
		$this->use_files($view, $files, $config);
	}
}
