<?php

Core::load('CMS.ORM');

class CMS_Stockroom_Entity extends CMS_ORM_Entity implements Core_ModuleInterface
{
	protected $rep = false;
	protected $installed_version = false;
	protected $hashes = array();
	protected $manager = null;

	public function setup()
	{
		$this->manager = CMS_Stockroom::manager();
		return parent::setup();
	}

	public function init_from($o)
	{
		$this->rep = $this->manager->current_repository();
		foreach ($o as $field => $value) {
			if ($field == 'install_only' || $field == 'not_install_if_exists') {
				$m = array();
				foreach ($value as $s) {
					$s = preg_replace('{^www/}', CMS::www() . '/', $s);
					$m[$s] = $s;
				}
				$value = $m;
			}
			$this[$field] = $value;
		}
		return $this;
	}

	/**
	 * @param string $file
	 * @param string $from
	 * @param string $to
	 * @return bool
	 */
	public function can_update_file($file, $from, $to)
	{
		$this->load_info();
		if (isset($this['install_only']) && is_array($this['install_only'])) {
			if (isset($this['install_only'][$file])) {
				return false;
			}
			foreach ($this['install_only'] as $f) {
				$f = trim($f);
				if ($f != '' && $f[strlen($f) - 1] == '/') {
					if (strpos($file, $f) === 0) {
						return !IO_FS::exists($to);
					}
				}
			}
		}
		return true;
	}

	public function install_temp_dir()
	{
		return CMS::temp_dir() . '/stockroom/' . $this->code();
	}

	public function code()
	{
		return $this->rep . $this->id();
	}

	public function save_info_file()
	{
		$title = $this['title'];
		$version = $this['version'];
		$content = "component {$title}\nversion {$version}\n";
		foreach ($this->hashes as $file => $hash) {
			$content .= "hash {$hash} {$file}\n";
		}
		file_put_contents($this->info_path(), $content);
		CMS::chmod_file($this->info_path());
	}

	public function info_path()
	{
		$dir = $this->manager->get_info_dir();
		if (!IO_FS::exists($dir)) {
			CMS::mkdirs($dir);
		}
		$code = $this->code();
		return "{$dir}/{$code}";
	}

	public function hashes()
	{
		$this->load_info();
		return $this->hashes;
	}

	public function load_info()
	{
		if (!$this->installed_version) {
			if ($content = $this->load_info_file()) {
				foreach (explode("\n", $content) as $line) {
					$line = trim($line);
					if ($m = Core_Regexps::match_with_results('{^version(.+)$}', $line)) {
						$this->installed_version = trim($m[1]);
					} elseif ($m = Core_Regexps::match_with_results('{^hash\s+([^\s]+)\s+(.+)$}', $line)) {
						$hash = trim($m[1]);
						$file = trim($m[2]);
						$this->hashes[$file] = $hash;
					}
				}
			}
		}
	}

	public function load_info_file()
	{
		$path = $this->info_path();
		if (IO_FS::exists($path)) {
			return file_get_contents($path);
		}
		return false;
	}

	public function set_installed_hash($file, $hash)
	{
		$this->load_info();
		return $this->hashes[$file] = $hash;
	}

	/**
	 * @param string $file
	 * @return bool
	 */
	public function file_was_changed($file)
	{
		$installed = $this->get_installed_hash($file);
		if (!$installed) {
			return false;
		}
		$existed = $this->get_existed_hash($file);
		if (!$existed) {
			return false;
		}
		return $installed != $existed;
	}

	public function get_installed_hash($file)
	{
		$this->load_info();
		if (isset($this->hashes[$file])) {
			return $this->hashes[$file];
		}
		return false;
	}

	public function get_existed_hash($file)
	{
		$path = "../$file";
		if (!IO_FS::exists($path)) {
			return false;
		}
		return md5_file($path);
	}

	public function installed()
	{
		$this->load_info();
		return $this->installed_version;
	}

	public function download()
	{
		$agent = Net_HTTP::Agent($this->manager->get_curl_options())->act_as_browser();
		return $agent->send($this['download']);
	}

	public function is_loaded()
	{
		$this->load_info();
		if ($this->coreload) {
			if (preg_match('!Core::load\((.*)\)!', $this->coreload, $matches)) {
				$class = Core_Types::class_name_for(trim($matches[1], '\'"'));
				try {
					return class_exists($class, false);
				} catch (Exception $e) {
					// in that case errors on lazy loading
					return false;
				}
			}
		}
		return false;
	}

	public function delete()
	{
		$this->load_info();
		foreach ($this->hashes as $file => $hash) {
			$path = '../' . $file;
			$this->manager->alter_package_files($file, $path, $this);
			IO_FS::rm($path);
			IO_FS::rm("$path.upd");
		}
		IO_FS::rm($this->info_path());
		return true;
	}

	public function tag_exists($tag_name)
	{
		foreach ($this->tags as $id => $tag) {
			if (strtolower($tag) == $tag_name) {
				return true;
			}
		}
		return false;
	}

	public function is_frontend()
	{
		return $this->tag_exists('frontend');
	}

	public function composer_exists()
	{
		$files = array_keys($this->hashes);
		foreach ($files as $file) {
			if (Core_Strings::contains($file, 'composer.json')) {
				return true;
			}
		}
		return false;
	}

}