<?php
/**
 * Cache.Backend.MemCache
 *
 * MemCache кэширование
 *
 * @package Cache\Backend\MemCached
 * @version 0.2.0
 */

/**
 * @package Cache\Backend\MemCached
 */
class Cache_Backend_MemCached implements Core_ModuleInterface
{

	const VERSION = '0.2.0';
	const DEFAULT_TIMEOUT = 60;

	/**
	 * Инициализация модуля
	 *
	 */
	static public function initialize()
	{
		if (!extension_loaded('memcached')) {
			throw new Cache_Exception('No memcached php-module');
		}
	}

	/**
	 * Фабричный метод, возвращает объект класса Cache.Backend.MemCached.Backend
	 *
	 * @param string $dsn
	 * @param int    $timeout
	 *
	 * @return Cache_Backend_MemCached_Backend
	 */
	public function Backend($dsn, $timeout = Cache_Backend_MemCached::DEFAULT_TIMEOUT)
	{
		return new Cache_Backend_MemCached_Backend($dsn, $timeout);
	}

}

/**
 * Класс реализующий memcache кэширование
 *
 * @package Cache\Backend\MemCache
 */
class Cache_Backend_MemCached_Backend extends Cache_Backend
{

	private $memcached;

	/**
	 * Конструктор
	 *
	 * @param string $dsn
	 * @param int    $timeout
	 */
	public function __construct($dsn, $timeout = Cache_Backend_MemCached::DEFAULT_TIMEOUT)
	{
		$m1 = Core_Regexps::match_with_results('{^memcached://([^:]+):?(\d+)?}', $dsn);
		if (!$m1) {
			throw new Cache_BadDSNException($dsn);
		}
		$this->memcached = new Memcached;
		if (!$this->memcached->addServer($m1[1], Core::if_null($m1[2], 11211))) {
			throw new Cache_Exception('Could not connect');
		}
		$this->timeout = $timeout;
	}

	public function get_all_keys()
	{
/*		$list = array(); 
		$slabs_by_server = $this->memcached->getExtendedStats('slabs'); 
		foreach($slabs_by_server as $server => $slabs) {
			if (empty($slabs)) {
				continue;
			}
			foreach($slabs AS $slab_id => $slab_meta) {
				if (!is_int($slab_id) || !is_array($slab_meta)) {
					continue;
				}
				$cdump = $this->memcached->getExtendedStats('cachedump', (int) $slab_id, 1000);
				if ($cdump && isset($cdump[$server]) && is_array($cdump[$server]) && !isset($cdump[$server]['CLIENT_ERROR'])) {
					$keys = array_keys($cdump[$server]);
					if ($keys) {
						$list = array_merge($list, $keys);
					}
				}
			} 
		}*/
		return $this->memcached->getAllKeys();
	}

	/**
	 * Закрывает соединение с memcache
	 *
	 */
	protected function close()
	{
		$this->memcached->quit();
	}

	/**
	 * Возвращает значение по ключу, если значение не установлено возвращает $default
	 *
	 * @param string $key
	 * @param        $default
	 *
	 * @return mixed
	 */
	public function get($key, $default = null)
	{
		$res = Core::if_false($this->memcached->get($key), $default);
		Events::call('cache.get', $key, $default, $res);
		return $res;
	}

	/**
	 * Устанавливает значение по ключу
	 *
	 * @param string $key
	 * @param        $value
	 * @param int    $timeout
	 *
	 * @return boolean
	 */
	public function set($key, $value, $timeout = null)
	{
		$timeout = Core::if_null($timeout, $this->timeout);
		Events::call('cache.set', $key, $value, $timeout);
		return $this->memcached->set($key, $value, false, $timeout);
	}

	/**
	 * Удалят значение из кэша
	 *
	 * @param string $key
	 *
	 * @return boolean
	 */
	public function delete($key)
	{
		Events::call('cache.delete', $key);
		return $this->memcached->delete($key);
	}

	/**
	 * Проверяет есть ли занчение с ключом $key в кэше
	 *
	 * @param string $key
	 *
	 * @return boolean
	 */
	public function has($key)
	{
		Events::call('cache.has', $key);
		return (boolean)$this->memcached->get($key);
	}

	/**
	 * Инвалидирует кэш
	 *
	 */
	public function flush()
	{
		Events::call('cache.flush');
		Events::call('cache.delete', $s = '*');
		return $this->memcached->flush();
	}

}

