<?php
/**
 * Работа с файловой системой
 *
 * Модуль реализует набор классов для работы с файлами и каталогами. Для файлов и
 * каталогов реализованы соответствующие объектные представления. Также реализован
 * объектный интерфейс для запроса списка объектов файловой системы по определенным
 * критериям.
 *
 * @package IO\FS
 * @version 0.2.2
 */
Core::load('Time', 'IO.Stream', 'MIME', 'Events', 'Config');

/**
 * Класс модуля
 *
 * Содержит набор фабричных методов для создания экземпляров классов модуля, а также
 * процедурный интерфейс для работы с файлами.
 *
 * @package IO\FS
 */
class IO_FS implements Core_ConfigurableModuleInterface
{
	const VERSION = '0.2.2';

	static protected $options = array(
		'dir_mod' => 0775,
		'file_mod' => 0664,
		'dir_own' => false,
		'file_own' => false,
		'dir_grp' => false,
		'file_grp' => false,
	);

	/**
	 * Инициализация
	 *
	 * @param array $options
	 */
	static public function initialize(array $options = array())
	{
		$opts = isset(Config::all()->iofs) ? (array)Config::all()->iofs : array();
		self::options($opts);
		self::options($options);
	}

	/**
	 * Устанавливает опции модуля
	 *
	 * @param array $options
	 *
	 * @return mixed
	 */
	static public function options(array $options = array())
	{
		if (count($options)) {
			Core_Arrays::update(self::$options, $options);
		}
		return self::$options;
	}

	/**
	 * Устанавливает/возвращает опцию модуля
	 *
	 * @param string $name
	 * @param        $value
	 *
	 * @return mixed
	 */
	static public function option($name, $value = null)
	{
		$prev = isset(self::$options[$name]) ? self::$options[$name] : null;
		if ($value !== null) {
			self::options(array($name => $value));
		}
		return $prev;
	}

	/**
	 * Создает объект класса IO.FS.File
	 *
	 * @param string $path
	 *
	 * @return IO_FS_File
	 */
	static public function File($path)
	{
		return $path instanceof IO_FS_File ? $path : new IO_FS_File($path);
	}

	/**
	 * Создает объект класса IO.FS.FileStream
	 *
	 * @param string $path
	 * @param string $mode
	 *
	 * @return IO_FS_FileStream
	 */
	static public function FileStream($path, $mode = IO_Stream::DEFAULT_OPEN_MODE)
	{
		return new IO_FS_FileStream($path, $mode);
	}

	/**
	 * Создает объект класса IO.FS.Stat
	 *
	 * @param  $object
	 *
	 * @return IO_FS_Stat
	 */
	static public function Stat($object)
	{
		return new IO_FS_Stat($object);
	}

	/**
	 * Создает объект класса IO.FS.Dir
	 *
	 * @param string|IO_FS_Dir $path
	 *
	 * @return IO_FS_Dir
	 */
	static public function Dir($path = '.')
	{
		return $path instanceof IO_FS_Dir ? $path : new IO_FS_Dir($path);
	}

	/**
	 * Создает объект  класса IO.FS.Path
	 *
	 * @param string $path
	 *
	 * @return IO_FS_Path
	 */
	static public function Path($path)
	{
		return new IO_FS_Path($path);
	}

	/**
	 * Создает объект класса IO.FS.Query
	 *
	 * @return IO_FS_Query
	 */
	static public function Query()
	{
		return new IO_FS_Query();
	}

	/**
	 * Возвращает объект классов IO.FS.Dir или IO.FS.File по заданному пути
	 *
	 * @param string|IO_FS_FSObject $path
	 *
	 * @return IO_FS_Dir|IO_FS_File|null
	 */
	static public function file_object_for($path)
	{
		$path = (string)$path;
		if (!self::exists($path)) {
			return null;
		}
		return self::is_dir($path) ? self::Dir($path) : self::File($path);
	}

	/**
	 * Возвращает объект класса IO.FS.Dir, соответствующий текущему каталогу.
	 *
	 * @return IO_FS_Dir
	 */
	static public function pwd()
	{
		return self::Dir(getcwd());
	}

	/**
	 * Переходит в указанный каталог и возвращает соответствующий объект класса IO.FS.Dir
	 *
	 * @param string $path
	 *
	 * @return IO_FS_Dir|null
	 */
	static public function cd($path)
	{
		return chdir($path) ? self::Dir(getcwd()) : null;
	}

	/**
	 * Создает каталог и возвращает соответствующий объект класса IO.FS.Dir
	 *
	 * @param string   $path
	 * @param null|int $mode
	 * @param bool     $recursive
	 *
	 * @return IO_FS_Dir|null
	 */
	static public function mkdir($path, $mode = null, $recursive = true)
	{
		$mode = self::get_permision_for(null, $mode, 'mod', 'dir');
		$old = umask(0);
		$rs = (self::exists((string)$path) || mkdir((string)$path, $mode, $recursive)) ? self::Dir($path) : null;
		umask($old);
		return $rs;
	}

	/**
	 * @param      $path
	 * @param      $value
	 * @param      $type
	 * @param      $obj
	 *
	 * @return mixed
	 */
	static protected function get_permision_for($path, $value, $type, $obj = null)
	{
		if (!is_null($value)) {
			return $value;
		}
		$object = !empty($path) ? (is_dir($path) ? 'dir' : 'file') : $obj;
		return self::option("{$object}_{$type}");
	}

	/**
	 * Изменяет права доступа к файлу
	 *
	 * @param string $file
	 * @param int    $mode
	 *
	 * @return boolean
	 */
	static public function chmod($file, $mode = null)
	{
		$mode = self::get_permision_for($file, $mode, 'mod');
		return $mode ? @chmod((string)$file, $mode) : false;
	}

	public static function chmod_recursive($path, $mode = null)
	{
		$object = self::file_object_for($path);
		if ($object->exists()) {
			$object->chmod($mode);
			if (self::is_dir($object->path)) {
				foreach (self::Query()->recursive(true, true)->apply_to($object) as $nested) {
					$nested->chmod($mode);
				}
			}
		}
	}

	/**
	 * Изменяет владельца файла
	 *
	 * @param      $file
	 * @param      $owner
	 *
	 * @return bool
	 */
	static public function chown($file, $owner = null)
	{
		$owner = self::get_permision_for($file, $owner, 'own');
		return $owner ? @chown((string)$file, $owner) : false;
	}

	/**
	 * Изменяет группу файла
	 *
	 * @param      $file
	 * @param      $group
	 *
	 * @return bool
	 */
	static function chgrp($file, $group = null)
	{
		$group = self::get_permision_for($file, $group, 'grp');
		return $group ? @chgrp((string)$file, $group) : false;
	}

	/**
	 * Удаляет файл
	 *
	 * @param string $path
	 *
	 * @return boolean
	 */
	static public function rm($path)
	{
		$obj = self::file_object_for($path);
		return $obj ? $obj->rm() : false;
	}

	/**
	 * Удаляет файл
	 *
	 * @param string $path
	 *
	 * @return boolean
	 */
	static public function clear_dir($path)
	{
		if (!self::is_dir($path)) {
			return false;
		}
		$dir = self::Dir($path);
		$rc = true;
		foreach ($dir as $resource) {
			$rc = $resource->rm() && $rc;
		}
		return $rc;
	}

	/**
	 * Создает вложенные каталоги
	 *
	 * @param string $path
	 * @param int    $mode
	 *
	 * @return boolean
	 */
	static public function make_nested_dir($path, $mode = null)
	{
		return self::mkdir($path, $mode, true);
	}

	/**
	 * Проверяет существование файла или каталога
	 *
	 * @param string $path
	 *
	 * @return boolean
	 */
	static public function exists($path)
	{
		return file_exists((string)$path);
	}

	/**
	 * Проверяет, является ли файловый объект с заданным путем каталогом
	 *
	 * @param string $path
	 *
	 * @return boolean
	 */
	static public function is_dir($path)
	{
		return is_dir((string)$path);
	}

	/**
	 * Перемещает файл
	 *
	 * @param string $from
	 * @param string $to
	 *
	 * @return boolean
	 */
	static public function mv($from, $to)
	{
		$obj = self::file_object_for($from);
		return $obj ? $obj->move_to($to) : false;
	}

	/**
	 * Копирует файл
	 *
	 * @param string $from
	 * @param string $to
	 *
	 * @return boolean
	 */
	static public function cp($from, $to)
	{
		$obj = self::file_object_for($from);
		return $obj ? $obj->copy_to($to) : false;
	}

}

/**
 * Базовый класс исключений модуля
 *
 * @package IO\FS
 */
class IO_FS_Exception extends IO_Exception
{
}

/**
 * Класс исключений для ошибок получения информации о файловом объекте
 *
 * @package IO\FS
 */
class IO_FS_StatException extends IO_FS_Exception
{
	protected $object;

	/**
	 * Конструктор
	 *
	 * @param  $object
	 */
	public function __construct($object)
	{
		parent::__construct("Can't stat object" . ((string)($this->object = $object)));
	}

}

/**
 * Объектное представление пути в файловой системе
 *
 * Объект представляет собой объектную обертку над встроенной функцией pathinfo().
 * Соответственно, он обеспечивает доступ к следующим свойствам:
 *
 * @package IO\FS
 *
 * @property $dirname   путь к файлу
 * @property $basename  базовое имя файла
 * @property $extension расширение
 * @property $filename  имя файла
 */
class IO_FS_Path implements Core_PropertyAccessInterface
{
	protected $info = array();

	/**
	 * Конструктор
	 *
	 * @param string $path
	 */
	public function __construct($path)
	{
		$this->info = @pathinfo((string)$path);
	}

	/**
	 * Возвращает значение свойства
	 *
	 * @param string $property
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 */
	public function __get($property)
	{
		if (isset($this->info[$property])) {
			return $this->info[$property];
		} elseif ($property == 'extension') {
			return '';
		} else {
			throw new Core_MissingPropertyException($property);
		}
	}

	/**
	 * Устанавливает значение свойства
	 *
	 * @param string $property
	 * @param        $value
	 *
	 * @return mixed
	 * @throws Core_ReadOnlyObjectException
	 */
	public function __set($property, $value)
	{
		throw new Core_ReadOnlyObjectException($this);
	}

	/**
	 * Проверяет установку значения свойства
	 *
	 * @param string $property
	 *
	 * @return boolean
	 */
	public function __isset($property)
	{
		return isset($this->info[$property]);
	}

	/**
	 * Удаляет свойство
	 *
	 * @param string $property
	 *
	 * @throws Core_ReadOnlyObjectException
	 */
	public function __unset($property)
	{
		throw new Core_ReadOnlyObjectException($this);
	}

}

/**
 * Объектное представление метаинформации об объекте файловой системы
 *
 * Объект предоставляет ту же информацию, что и встроенные функции fstat() и stat(), однако
 * делает это более удобным образом: при создании объекта может быть использован как путь
 * к файлу, так и файловый ресурс, все свойства, содержащие дату, возвращаются в виде
 * объектов класса Time.DateTime.
 *
 * @package IO\FS
 *
 * @property               $nlink   количество ссылок
 * @property               $uid     userid владельца (В Windows это всегда будет 0)
 * @property               $gid     groupid владельца (В Windows это всегда будет 0)
 * @property               $size    размер в байтах
 * @property Time_DateTime $atime
 * @property Time_DateTime $mtime
 * @property Time_DateTime $ctime
 */
class IO_FS_Stat implements Core_PropertyAccessInterface
{
	protected $stat = array();

	public function __construct($object)
	{
		if (!$stat = Core_Types::is_resource($object) ? @fstat($object) : @stat((string)$object)) {
			throw new IO_FS_StatException($object);
		}

		foreach ($stat as $k => $v) {
			switch ($k) {
				case 'atime':
				case 'mtime':
				case 'ctime':
					$this->stat[$k] = Time::DateTime($v);
					break;
				default:
					$this->stat[$k] = $v;
			}
		}
	}

	/**
	 * Возвращает значние свойства
	 *
	 * @param string $property
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 */
	public function __get($property)
	{
		if (isset($this->stat[$property])) {
			return $this->stat[$property];
		} else {
			throw new Core_MissingPropertyException($property);
		}
	}

	/**
	 * Устанавливает значение свойства
	 *
	 * @param string $property
	 * @param        $value
	 *
	 * @return mixed
	 * @throws Core_ReadOnlyObjectException
	 */
	public function __set($property, $value)
	{
		throw new Core_ReadOnlyObjectException($this);
	}

	/**
	 * Проверяет установку значения свойства
	 *
	 * @param string $property
	 *
	 * @return boolean
	 */
	public function __isset($property)
	{
		return isset($this->stat[$property]);
	}

	/**
	 * Удаляет свойство
	 *
	 * @param string $property
	 *
	 * @throws Core_ReadOnlyObjectException
	 */
	public function __unset($property)
	{
		throw new Core_ReadOnlyObjectException($this);
	}

}

/**
 * Базовый класс объектов файловой системы
 *
 * Предназначен для использования в качестве базового класса для более специфичных классов,
 * представляющих файлы и каталоги. Каждый объект класса характеризуется своим именем и
 * метаданными, содержащимися в виде объекта IO.FS.Stat, загружаемого по требованию.
 * Класс также реализует набор операций, в равной степени применимых к файлам и каталогам.
 *
 * @package IO\FS
 *
 * @property string $path      путь к файлу
 * @property string $dir_name  имя каталога
 * @property string $name      имя файла
 * @property string $real_path реальный путь к файлу (с раскрытыми ., .., и т.д.)
 * @property string $stat      метаданные в виде объекта класса IO.FS.Stat
 */
class IO_FS_FSObject implements
	Core_StringifyInterface,
	Core_EqualityInterface
{
	protected $path;
	protected $stat;

	/**
	 * Конструктор
	 *
	 * @param string $path
	 */
	public function __construct($path)
	{
		$this->path = $path;
	}

	/**
	 * Изменяет права доступа к файловому объекту
	 *
	 * @param int $mode
	 *
	 * @return boolean
	 */
	public function chmod($mode = null)
	{
		$this->stat = null;
		return IO_FS::chmod($this->path, $mode);
	}

	/**
	 * Изменяет владельца файлового объекта
	 *
	 * @param  $owner
	 *
	 * @return boolean
	 */
	public function chown($owner = null)
	{
		$this->stat = null;
		return IO_FS::chown($this->path, $owner);
	}

	public function chgrp($group = null)
	{
		$this->stat = null;
		return IO_FS::chgrp($this->path, $group);
	}

	public function set_permission($mode = null, $owner = null, $group = null)
	{
		$this->chmod($mode);
		$this->chown($owner);
		$this->chgrp($group);
		return $this;
	}

	public function exists()
	{
		return file_exists($this->path);
	}

	/**
	 * Возвращает значение свойства
	 *
	 * @param string $property
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 */
	public function __get($property)
	{
		switch ($property) {
			case 'path':
				return $this->path;
			case 'dir_name':
			case 'dirname':
				return @dirname($this->path);
			case 'name':
				return @basename($this->path);
			case 'real_path':
				return @realpath($this->path);
			case 'stat':
				if (!$this->stat) {
					$this->stat = IO_FS::Stat($this->path);
				}
				return $this->stat;
			default:
				throw new Core_MissingPropertyException($property);
		}
	}

	/**
	 * Устанавливает значение свойства
	 *
	 * @param string $property
	 * @param        $value
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 * @throws Core_ReadOnlyPropertyException
	 */
	public function __set($property, $value)
	{
		switch ($property) {
			case 'path':
			case 'dir_name':
			case 'name':
			case 'real_path':
			case 'stat':
				throw new Core_ReadOnlyPropertyException($property);
			default:
				throw new Core_MissingPropertyException($property);
		}
	}

	/**
	 * Проверяет установку свойства объекта
	 *
	 * @param string $property
	 *
	 * @return boolean
	 */
	public function __isset($property)
	{
		switch ($property) {
			case 'path':
			case 'dir_name':
			case 'name':
			case 'real_path':
			case 'stat':
				return true;
			default:
				return false;
		}
	}

	/**
	 * Удаляет свойство объекта
	 *
	 * @param string $property
	 *
	 * @throws Core_MissingPropertyException
	 * @throws Core_UndestroyablePropertyException
	 */
	public function __unset($property)
	{
		switch ($property) {
			case 'path':
			case 'dir_name':
			case 'name':
			case 'real_path':
			case 'stat':
				throw new Core_UndestroyablePropertyException($property);
			default:
				throw new Core_MissingPropertyException($property);
		}
	}

	/**
	 * @return string
	 */
	public function as_string()
	{
		return $this->real_path;
	}

	/**
	 * @return string
	 */
	public function __toString()
	{
		return $this->real_path;
	}

	/**
	 * @param  $to
	 *
	 * @return boolean
	 */
	public function equals($to)
	{
		return $to instanceof self && $this->real_path == $to->real_path;
	}

}

/**
 * Объектное представление файла
 *
 * Файл представляет собой файловый объект, для которого могут быть получены метаданные и
 * открыт файловый поток. Объект файла предоставляет также возможность получения информации
 * о MIME-типе, соответствующем файлу, с использованием модуля MIME.
 * Объект может работать как итератор, в этом случае соответствующий файл открывается на
 * чтение и для него создается соответствующий поток.
 *
 * @package IO\FS
 *
 * @property  $stream       ассоциированный поток
 * @property  $size         размер файла
 * @property  $mime_type    MIME-тип файла в виде объекта
 * @property  $content_type MIME-тип файла в виде строки
 */
class IO_FS_File extends IO_FS_FSObject implements
	Core_PropertyAccessInterface,
	IteratorAggregate
{
	protected $mime_type;
	protected $stream;

	/**
	 * Деструктор
	 *
	 * TODO: Тут наверно __destruct всеже?
	 */
	public function __destroy()
	{
		$this->close();
		parent::__destroy();
	}

	public function create()
	{
		return $this->update('', null);
	}

	/**
	 * Создает поток класса IO.FS.FileStream, соответствующий файлу
	 *
	 * @param string $mode
	 *
	 * @return IO_FS_FileStream
	 */
	public function open($mode = IO_Stream::DEFAULT_OPEN_MODE)
	{
		return ($this->stream && $this->stream->id) ? $this->stream : $this->stream = new IO_FS_FileStream($this->path, $mode);
	}

	/**
	 * Закрывает поток, ассоциированный с файловым объектом
	 *
	 * @return IO_FS_File
	 */
	public function close()
	{
		$this->stream = null;
		return $this;
	}

	/**
	 * Возвращает все содержимое файла в виде строки
	 *
	 * @param null $use_include_path
	 * @param null $context
	 * @param int  $offset
	 * @param null $maxlen
	 *
	 * @return string
	 */
	public function load($use_include_path = null, $context = null, $offset = 0, $maxlen = null)
	{
		return file_get_contents($this->path, $use_include_path, $context, $offset, $maxlen ? $maxlen : $this->size);
	}

	/**
	 * Эффективно записывает блок данных или строку в начало файла
	 *
	 * @param string $data
	 * @param int    $flags
	 *
	 * @return mixed
	 */
	public function update($data, $flags = 0)
	{
		$res = file_put_contents($this->path, $data, (int)$flags);
		$this->set_permission();
		return $res;
	}

	/**
	 * Добавляет данные в конец файла
	 *
	 * @param string $data
	 * @param int    $flags
	 *
	 * @return mixed
	 */
	public function append($data, $flags = 0)
	{
		$res = file_put_contents($this->path, $data, FILE_APPEND | $flags);
		$this->set_permission();
		return $res;
	}

	/**
	 * Перемещает файл
	 *
	 * @param string $destination
	 *
	 * @return IO_FS_File
	 */
	public function move_to($destination)
	{
		if ($this->stream && $this->stream->id) {
			return null;
		}

		if (rename($this->path, $destination = $this->fix_destination($destination))) {
			$this->path = $destination;
			$this->stat = null;
			return $this;
		} else {
			return null;
		}
	}

	/**
	 * Копирует файл
	 *
	 * @param string $destination
	 *
	 * @return IO_FS_File
	 */
	public function copy_to($destination)
	{
		if ($this->stream && $this->stream->id) {
			return null;
		}
		$copied = copy($this->path, $destination = $this->fix_destination($destination)) ? IO_FS::File($destination) : null;
		return $copied;
	}

	public function rm()
	{
		return unlink($this->path);
	}

	/**
	 * Создает итератор для файлового объекта
	 */
	public function getIterator()
	{
		return $this->open()->getIterator();
	}

	/**
	 * Возвращает значение свойства
	 *
	 * @param string $property
	 *
	 * @return mixed
	 */
	public function __get($property)
	{
		switch ($property) {
			case 'stream':
				return $this->stream;
			case 'size':
				return @filesize($this->path);
			case 'mime_type':
				return $this->get_mime_type();
			case 'content_type':
				return $this->get_mime_type()->type;
			default:
				return parent::__get($property);
		}
	}

	/**
	 * Устанавливает значение свойства
	 *
	 * @param string $property
	 * @param        $value
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 * @throws Core_ReadOnlyPropertyException
	 */
	public function __set($property, $value)
	{
		switch ($property) {
			case 'stream':
			case 'size':
			case 'mime_type':
			case 'content_type':
				throw new Core_ReadOnlyPropertyException($property);
			default:
				return parent::__set($property, $value);
		}
	}

	/**
	 * Проверяет установку свойства
	 *
	 * @param string $property
	 *
	 * @return boolean
	 */
	public function __isset($property)
	{
		switch ($property) {
			case 'stream':
				return isset($this->stream);
			case 'size':
			case 'mime_type':
			case 'content_type':
				return true;
			default:
				return parent::__isset($property);
		}
	}

	/**
	 * Удаляет свойство
	 *
	 * @param string $property
	 *
	 * @throws Core_MissingPropertyException
	 * @throws Core_UndestroyablePropertyException
	 */
	public function __unset($property)
	{
		if ($this->__isset($property)) {
			throw new Core_UndestroyablePropertyException($property);
		} else {
			parent::__unset($property);
		}
	}

	/**
	 * Возвращает MIME-тип, соответствующий файлу, в виде объекта класса MIME.Type
	 *
	 * @return MIME_Type
	 */
	private function get_mime_type()
	{
		return $this->mime_type ? $this->mime_type : $this->mime_type = MIME::type_for_file($this);
	}

	/**
	 * Корректирует новый путь к файлу для операций копирования и перемещения
	 *
	 * @param $destination
	 *
	 * @return string
	 */
	private function fix_destination($destination)
	{
		if (is_dir($destination)) {
			return (Core_Strings::ends_with($destination, '/') ? $destination : $destination . '/') . @basename($this->path);
		}
		return $destination;
	}

}

/**
 * Файловый поток
 *
 * Расширяет базовый класс IO.Stream.NamedResourceStream поддержкой интерфейса
 * IO.Stream.SeekableInterface и реализацией метода truncate().
 *
 * @package IO\FS
 */
class IO_FS_FileStream extends IO_Stream_NamedResourceStream implements IO_Stream_SeekInterface
{
	/**
	 * Устанавливает текущую позицию в потоке
	 *
	 * @param int $offset
	 * @param int $whence
	 *
	 * @return int
	 */
	public function seek($offset, $whence = SEEK_SET)
	{
		return @fseek($this->id, $offset, $whence);
	}

	/**
	 * Возвращает текущую позицию в потоке
	 *
	 * @return int
	 */
	public function tell()
	{
		return @ftell($this->id);
	}

	/**
	 * Обрезает файл до заданной длины
	 *
	 * @param int $size
	 *
	 * @return boolean
	 */
	public function truncate($size = 0)
	{
		return @ftruncate($this->id, $size);
	}

}

/**
 * Каталог файловой системы
 *
 * Расширяет базовый класс файлового объекта следующими возможностями:
 * - получение объектного представления элементов каталога с помощью операции доступа по индексу
 * - итератор по элементам каталога
 * - овместная работа с объектами запроса списка файлов, позволяющих запрашивать
 *   списки элементов каталога, удовлетворяющих определенным критериям.
 *
 * @property $files итератор по содержимому каталога с настройками по умолчанию
 *
 * @package IO\FS
 */
class IO_FS_Dir extends IO_FS_FSObject implements
	Core_PropertyAccessInterface,
	Core_IndexedAccessInterface,
	IteratorAggregate
{
	/**
	 * Конструктор
	 *
	 * @param string $path
	 */
	public function __construct($path)
	{
		parent::__construct(rtrim($path, '/'));
	}

	/**
	 * Применяет к каталогу объект, содержащий условия запроса его содержимого
	 *
	 * @param IO_FS_Query $query
	 *
	 * @return IO_FS_Dir
	 */
	public function query(IO_FS_Query $query)
	{
		return $query->apply_to($this);
	}

	/**
	 * Возвращает значение свойства
	 *
	 * @param string $property
	 *
	 * @return mixed
	 */
	public function __get($property)
	{
		switch ($property) {
			case 'files':
				return $this->make_default_iterator();
			default:
				return parent::__get($property);
		}
	}

	/**
	 * Устанавливает значение свойства
	 *
	 * @param string $property
	 * @param        $value
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 * @throws Core_ReadOnlyObjectException
	 * @throws Core_ReadOnlyPropertyException
	 */
	public function __set($property, $value)
	{
		switch ($property) {
			case 'files':
				throw new Core_ReadOnlyObjectException($property);
			default:
				return parent::__set($property, $value);
		}
	}

	/**
	 * Проверяет установку значения свойства
	 *
	 * @param string $property
	 *
	 * @return boolean
	 */
	public function __isset($property)
	{
		switch ($property) {
			case 'files':
				return true;
			default:
				return parent::__isset($property);
		}
	}

	/**
	 * Удаляет свойство
	 *
	 * @param string $property
	 *
	 * @throws Core_MissingPropertyException
	 * @throws Core_UndestroyablePropertyException
	 */
	public function __unset($property)
	{
		if ($this->__isset($property)) {
			throw new Core_UndestroyablePropertyException($property);
		} else {
			parent::__unset($property);
		}
	}

	/**
	 * Возвращает итератор по содержимому каталога.
	 *
	 * @return IO_FS_DirIterator
	 */
	public function getIterator()
	{
		return $this->make_default_iterator();
	}

	/**
	 * Возвращает значение индексированного свойства
	 *
	 * @param mixed $index
	 *
	 * @return IO_FS_Dir|IO_FS_File|null
	 */
	public function offsetGet($index)
	{
		return IO_FS::file_object_for("{$this->path}/$index");
	}

	/**
	 * Устанавливает значение индексированного свойства
	 *
	 * @param string $index
	 * @param mixed  $value
	 *
	 * @throws Core_MissingIndexedPropertyException
	 * @throws Core_ReadOnlyIndexedPropertyException
	 */
	public function offsetSet($index, $value)
	{
		if ($this->offsetExists($index)) {
			throw new Core_ReadOnlyIndexedPropertyException($index);
		}
		throw new Core_MissingIndexedPropertyException($index);
	}

	/**
	 * Проверяет существование индексированного свойства
	 *
	 * @param string $index
	 *
	 * @return boolean
	 */
	public function offsetExists($index)
	{
		return IO_FS::exists("{$this->path}/$index");
	}

	/**
	 * Удаляет индексированное свойство
	 *
	 * @param string $index
	 *
	 * @throws Core_MissingIndexedPropertyException
	 * @throws Core_ReadOnlyIndexedPropertyException
	 */
	public function offsetUnset($index)
	{
		if (isset($this[$index])) {
			throw new Core_ReadOnlyIndexedPropertyException($index);
		}
		throw new Core_MissingIndexedPropertyException($index);
	}

	public function rm()
	{
		if (!$this->exists()) {
			return $this;
		}
		foreach ($this as $obj) {
			$obj->rm();
		}
		rmdir($this->path);
		return $this;
	}

	public function create()
	{
		return IO_FS::mkdir($this->path, null, true);
	}

	public function copy_to($dest)
	{
		$dest = IO_FS::Dir($dest);
		if (!$this->exists()) {
			return $dest;
		}
		$dest->rm()->create();
		foreach ($this as $obj) $obj->copy_to($dest->path . DIRECTORY_SEPARATOR . $obj->name);
		return $dest;
	}

	public function move_to($dest)
	{
		$dest = $this->copy_to($dest);
		$this->rm();
		return $dest;
	}

	/**
	 * Возвращает итератор по умолчанию
	 *
	 * @return IO_FS_DirIterator
	 */
	public function make_default_iterator()
	{
		return IO_FS::Query()->apply_to($this);
	}

}

/**
 * Условия выборки элементов каталога
 *
 * Временная реализация с минимальной функциональностью -- поиск по регулярному выражению
 * или шаблону с возможностью рекурсии. Планируется изменить реализацию для поддержки поиска
 * по датам, типам файлов и т.д.
 *
 * @package IO\FS
 *
 * @property $regexp    регулярное выражение
 * @property $recursive признак выполнения рекурсивного поиска
 */
class IO_FS_Query implements Core_PropertyAccessInterface
{
	const DEFAULT_REGEXP = '{.+}';

	protected $regexp = self::DEFAULT_REGEXP;
	protected $recursive = false;
	protected $self = false;

	/**
	 * Задает регулярное выражение для поиска
	 *
	 * @param string $regexp
	 *
	 * @return IO_FS_Query
	 */
	public function regexp($regexp)
	{
		$this->regexp = (string)$regexp;
		return $this;
	}

	/**
	 * Задает шаблон поиска с использованием * и ?
	 *
	 * @param string $wildcard
	 *
	 * @return IO_FS_Query
	 */
	public function glob($wildcard)
	{
		$this->regexp = '{' . Core_Strings::replace(Core_Strings::replace(Core_Strings::replace($wildcard, '.', '\.'), '?', '.'), '*', '.*') . '}';
		return $this;
	}

	/**
	 * Устанавливает флаг рекурсивного поиска в каталоге
	 *
	 * @param boolean $use_recursion
	 * @param bool    $self
	 *
	 * @return IO_FS_Query
	 */
	public function recursive($use_recursion = true, $self = false)
	{
		$this->recursive = (boolean)$use_recursion;
		$this->self = $self;
		return $this;
	}

	/**
	 * Возвращает итератор, соответствующий условиям поиска
	 *
	 * @param IO_FS_Dir $dir
	 *
	 * @return IO_FS_DirIterator
	 */
	public function apply_to(IO_FS_Dir $dir)
	{
		if ($this->recursive) {
			if (!$this->self) {
				return new RecursiveIteratorIterator(new IO_FS_DirIterator($dir, $this));
			} else {
				return new RecursiveIteratorIterator(new IO_FS_DirIterator($dir, $this), RecursiveIteratorIterator::SELF_FIRST);
			}
		} else {
			return new IO_FS_DirIterator($dir, $this);
		}
	}

	/**
	 * Проверяет, является ли поиск рекурсивным
	 *
	 * @return boolean
	 */
	public function is_recursive()
	{
		return $this->recursive;
	}

	/**
	 * Проверяет, соответствует ли заданный путь условиям поиска
	 *
	 * @param string $path
	 *
	 * @return boolean
	 */
	public function allows($path = '.')
	{
		if (IO_FS::is_dir($path) && $this->recursive) {
			return true;
		}

		return Core_Regexps::match($this->regexp, (string)$path);
	}

	/**
	 * Проверяет отсутствие сооответствия заданного пути условиям поиска
	 *
	 * @param string $path
	 *
	 * @return boolean
	 */
	public function forbids($path = '.')
	{
		return !$this->allows($path);
	}

	/**
	 * Возвращает значение свойства
	 *
	 * @param string $property
	 *
	 * @return mixed
	 * @throws Core_MissingPropertyException
	 */
	public function __get($property)
	{
		switch ($property) {
			case 'regexp':
			case 'recursive':
				return $this->$property;
			default:
				throw new Core_MissingPropertyException($property);
		}
	}

	/**
	 * Устанавливает значение свойства
	 *
	 * @param string $property
	 * @param        $value
	 *
	 * @return IO_FS_Query
	 * @throws Core_MissingPropertyException
	 */
	public function __set($property, $value)
	{
		switch ($property) {
			case 'regexp':
				$this->regexp = (string)$value;
				break;
			case 'recursive':
				$this->recursive = (boolean)$value;
				break;
			default:
				throw new Core_MissingPropertyException($property);
		}
		return $this;
	}

	/**
	 * Проверяет установку значения свойства
	 *
	 * @param string $property
	 *
	 * @return boolean
	 */
	public function __isset($property)
	{
		switch ($property) {
			case 'regexp':
			case 'recursive':
				return isset($this->$property);
			default:
				return false;
		}
	}

	/**
	 * Удаляет свойство
	 *
	 * @param string $property
	 *
	 * @throws Core_MissingPropertyException
	 * @throws Core_UndestroyablePropertyException
	 */
	public function __unset($property)
	{
		switch ($property) {
			case 'regexp':
			case 'recursive':
				throw new Core_UndestroyablePropertyException($property);
			default:
				throw new Core_MissingPropertyException($property);
		}
	}

}

/**
 * Итератор по содержимому каталога
 *
 * Предназначен для использования вместе с объектами класса IO.FS.Dir.
 *
 * @package IO\FS
 */
class IO_FS_DirIterator implements RecursiveIterator
{
	/** @var IO_FS_Dir|IO_FS_File|null|bool */
	protected $current = false;
	protected $query;
	protected $dir;
	protected $id;

	/**
	 * Конструктор
	 *
	 * @param IO_FS_Dir   $dir
	 * @param IO_FS_Query $query
	 */
	public function __construct(IO_FS_Dir $dir, IO_FS_Query $query = null)
	{
		$this->dir = $dir;
		$this->query = $query ? $query : IO_FS::Query();
	}

	/**
	 * Проверяет, является ли текущий элемент подкаталогом
	 *
	 * @return boolean
	 */
	public function hasChildren()
	{
		return ($this->query->is_recursive() && ($this->current instanceof IO_FS_Dir));
	}

	/**
	 * Возвращает итератор по подкаталогу
	 *
	 * @return IO_FS_DirIterator
	 */
	public function getChildren()
	{
		return new IO_FS_DirIterator($this->current, $this->query);
	}

	/**
	 * Возвращает текущий элемент
	 *
	 * @return IO_FS_FSObject
	 */
	public function current()
	{
		return $this->current;
	}

	/**
	 * Возвращает ключ текущего элемента
	 *
	 * @return string
	 */
	public function key()
	{
		return $this->current->path;
	}

	/**
	 * Сбрасывает итератор
	 *
	 */
	public function rewind()
	{
		if ($this->id) {
			@closedir($this->id);
		}
		rewinddir($this->id = @opendir($this->dir->path));
		$this->skip_to_next();
	}

	/**
	 * Переходит к следующему элементу итератора
	 *
	 */
	public function next()
	{
		return $this->skip_to_next();
	}

	/**
	 * Проверяет существование текущего элемента итератора
	 *
	 * @return boolean
	 */
	public function valid()
	{
		return $this->current ? true : false;
	}

	/**
	 * Переходит к очередному элементу, удовлетворяющему условиям поиска
	 *
	 */
	protected function skip_to_next()
	{
		do {
			$name = readdir($this->id);
			$path = $this->dir->path . '/' . $name;
		} while ($name !== false && ($name == '.' || $name == '..' || $this->query->forbids($path = $this->dir->path . '/' . $name)));
		if ($name !== false) {
			$this->current = IO_FS::file_object_for($path);
		} else {
			@closedir($this->id);
			$this->id = null;
			$this->current = null;
		}
		return $this->current;
	}
}
