<?php
/**
 * @package Tree\DB\MP
 */

// Implementation of Materialized Path

Core::load('Tree.DB');

class Tree_DB_MP implements Core_ModuleInterface
{

	const VERSION = '0.1.0';

	static public function schema(&$data)
	{
		if (!isset($data['columns']['path'])) {
			$data['columns']['path'] = array('type' => 'varchar', 'length' => 255, 'default' => '', 'not null' => true);
		}
		if (!isset($data['columns']['path1'])) {
			$data['columns']['path1'] = array('type' => 'varchar', 'length' => 255, 'default' => '', 'not null' => true);
		}
		if (!isset($data['columns']['level'])) {
			$data['columns']['level'] = array('type' => 'int', 'default' => '0', 'not null' => true);
		}
		if (!isset($data['columns']['parent_id'])) {
			$data['columns']['parent_id'] = array('type' => 'int', 'default' => '0', 'not null' => true);
		}
		if (!isset($data['columns']['ord'])) {
			$data['columns']['ord'] = array('type' => 'int', 'default' => '0', 'not null' => true);
		}
		$table = $data['name'] . '_tree';
		$data['indexes']["idx_{$table}_path"] = array('columns' => array(array('path', 255)));
		$data['indexes']["idx_{$table}_path1"] = array('columns' => array(array('path1', 255)));
		$data['indexes']["idx_{$table}_level"] = array('columns' => array('level'));
		$data['indexes']["idx_{$table}_parent_id"] = array('columns' => array('parent_id'));
		$data['indexes']["idx_{$table}_ord"] = array('columns' => array('ord'));
	}
}

/**
 * Class Tree_DB_MP_Mapper
 *
 * @method $this descendants($item)
 */
class Tree_DB_MP_Mapper extends Tree_DB_AbstractMapper
{
	use Tree_DB_MP_MapperTrait;
}

trait Tree_DB_MP_MapperTrait
{
	protected function setup()
	{
		parent::setup();
		$this->columns('path', 'path1', 'level', 'ord', 'parent_id')
			->option('delimiter', '.');

		return $this;
	}

	public function generate()
	{
		$this->update_all(array('path' => '', 'level' => 0));
		foreach ($this as $row) {
			$this->generate_item($row);
		}
		return $this;
	}

	public function generate_item($row)
	{
		$path = $this->generate_item_path($row);
		$row->path = $path;
		$row->level = $row->calculate_level();
		$row->manual_setup = true;
		$row->update();
	}

	public function generate_item_path($item)
	{
		$parent = $item->parent;
		if (!$parent) {
			return $item->id();
		}
		return $this->generate_item_path($parent) . $this->option('delimiter') . $item->id();
	}

	protected function map_sort_tree($add = '')
	{
		return $this->order_by('level, ord' . (!empty($add) ? ",$add" : ''));
	}

	protected function map_parent($item)
	{
		return $this->where('id = :id', $item->parent_id);
	}

	protected function map_children($item)
	{
		return $this->descendants($item)->where('level = :level', $item->level + 1);
	}

	protected function map_siblings($item, $include_self = false)
	{
		$m = $this->where('parent_id = :pid', $item->parent_id);
		if (!$include_self) {
			$m->where('id <> :cid', $item->id);
		}
		return $m;
	}

	protected function map_descendants($item)
	{
		$path_column = $this->path_column($item);
		return $this->where("$path_column LIKE :pp", $item->$path_column . $this->option('delimiter') . '%');
	}

	protected function path_column($item)
	{
		return empty($item->path1) ? 'path' : 'path1';
	}

	protected function map_full_tree()
	{
		return $this;
	}
}

class Tree_DB_MP_Entity extends DB_ORM_Entity
{
	use Tree_DB_EntityTrait;
	use Tree_DB_MP_EntityTrait;
}

trait Tree_DB_MP_EntityTrait
{

	public function setup()
	{
		parent::setup();
		$this->assign_attrs(array(
				'level' => 0,
				'path' => '',
				'path1' => '',
				'ord' => 0,
				'parent_id' => 0
			)
		);
		return $this;
	}

	public function row_set_parent_id($value)
	{
		$old_parent_id = $this->get('old_parent_id');
		if (empty($old_parent_id)) {
			$this->set('old_parent_id', $this->get('parent_id'));
		}
		$this->set('parent_id', $value);
		return $this;
	}

	public function set_parent_id($value)
	{
		return $this->row_set_parent_id($value);
	}

	public function set_path($value)
	{
		$dl = $this->mapper->option('delimiter');
		$path = !is_array($value) ? (string)$value : implode($dl, $value);
		$ids = explode($this->mapper->option('delimiter'), $path);
		$ids1 = array();
		while (mb_strlen($path) >= 255) {
			$ids1[] = array_pop($ids);
			$path = implode($dl, $ids);
		}
		$ids1 = array_reverse($ids1);
		$this->set('path', $path);
		$this->set('path1', implode($dl, $ids1));
	}

	public function get_path()
	{
		$path1 = $this->get('path1');
		if (!empty($path1)) {
			return $this->get('path') . $this->mapper->option('delimiter') . $path1;
		}
		return $this->get('path');
	}

	public function get_trail($include_self = true)
	{
		$ids = explode($this->mapper->options['delimiter'], $this->path);
		if (!$include_self) {
			$search = array_search($this->id(), $ids, false);
			if ($search !== false) {
				unset($ids[$search]);
			}
		}
		return empty($ids) ? array() : $this->mapper->spawn()->in('id', $ids)->order_by('level')->select();
	}

	public function calculate_level()
	{
		$path_ids = explode($this->mapper->option('delimiter'), $this->path);
		return max(count($path_ids) - 1, 0);
	}

	public function after_insert()
	{
		$parent = $this->get_parent();
		if ($parent) {
			$this->path = $parent->path . $this->mapper->option('delimiter') . $this->id();
		} else {
			$this->path = $this->id();
		}
		$this->level = $this->calculate_level();
		$this->old_parent_id = $this->parent_id;
		$this->update(array('path', 'path1', 'level'));
		return parent::after_insert();
	}

	public function after_update()
	{
		if (!$this->manual_setup && $this->old_parent_id != $this->parent_id) {
			$parent = $this->get_parent();

			$old_descendants = $this->mapper->descendants($this);
			$old_level = $this->level;
			$old_path = $this->path;
			$ids = array();
			foreach ($this->get_descendants() as $d)
				$ids[] = $d->id();
			// Very bad case:
			// ----------
			if ($parent && in_array($parent->id(), $ids)) {
				$parent->parent_id = $this->old_parent_id;
				$parent->update();
				return $this->after_update();
			}
			// ----------
			if ($parent) {
				$this->path = $parent->path . $this->mapper->option('delimiter') . $this->id();
				$this->level = $this->calculate_level();
			} else {
				$this->path = $this->id();
				$this->level = 0;
			}

			$this->old_parent_id = $this->parent_id;
			$this->update(array('path', 'path1', 'level'));

			$dist = $this->level - $old_level;
			$old_descendants->update_all(array(), array('level' => "level + $dist", 'path' => "REPLACE(path, '$old_path', '{$this->path}')"));
		}
		return parent::after_update();
	}

	public function after_delete()
	{
		$this->mapper->descendants($this)->delete_all();
		return parent::after_delete();
	}

}