<?php
/**
 * @package Tree\DB\MPT
 */

// Implementation of Materialized Path by separate table

Core::load('Tree.DB');

class Tree_DB_MPT implements Core_ModuleInterface
{

	const VERSION = '0.1.0';

	static public function schema(&$data, $create_join_table = true)
	{
		$table = $data['name'];
		if (!isset($data['columns']['parent_id'])) {
			$data['columns']['parent_id'] = array('type' => 'int', 'default' => '0', 'not null' => true);
			$data['indexes']["idx_{$table}_parent_id"] = array('columns' => array('parent_id'));
		}
		if ($create_join_table && !empty($data['name'])) {
			$table = $data['name'] . '_tree';
			$join_schema = array($table => array());
			self::schema_join($join_schema[$table], $table);
			DB_Schema::process($join_schema);
		}
	}

	static public function schema_join(&$data, $table = null)
	{
		$data = array_merge_recursive($data, array(
				'columns' => array(
					'id' => array('type' => 'serial'),
					'entity_id' => array('type' => 'int', 'default' => '0', 'not null' => true),
					'parent_id' => array('type' => 'int', 'default' => '0', 'not null' => true),
					'level' => array('type' => 'int', 'default' => '0', 'not null' => true),
					'ord' => array('type' => 'int', 'default' => '0', 'not null' => true)
				),
				'indexes' => array(
					array('type' => 'primary key', 'columns' => array('id')),
					"idx_{$table}_entity_id" => array('columns' => array('entity_id')),
					"idx_{$table}_parent_id" => array('columns' => array('parent_id')),
					"idx_{$table}_level" => array('columns' => array('level')),
					"idx_{$table}_ord" => array('columns' => array('ord')),
				)
			)
		);
	}
}

class Tree_DB_MPT_Mapper extends Tree_DB_AbstractMapper
{
	use Tree_DB_MPT_MapperTrait;
}

trait Tree_DB_MPT_MapperTrait
{

	public function generate()
	{
		$this->connection->execute('TRUNCATE ' . $this->tree_table());
		foreach ($this->spawn() as $row) {
			$this->generate_item($row);
			unset($row);
		}
		return $this;
	}

	public function generate_item($row)
	{
		$path = $this->generate_item_path($row);
		$ids = explode(',', $path);
		$level = count($ids) - 1;
		foreach ($ids as $pid) {
			if ($pid == $row->id()) {
				continue;
			}
			$this->get_tree_mapper()->make_entity(array('entity_id' => $row->id(), 'parent_id' => $pid, 'level' => $level))->insert();
		}
		if (count($ids) == 1) {
			$this->get_tree_mapper()->make_entity(array('entity_id' => $row->id(), 'parent_id' => 0, 'level' => 0))->insert();
		}
	}

	public function generate_item_path($item)
	{
		$parent = $item->parent;
		if (!$parent) {
			return $item->id();
		}
		return $this->generate_item_path($parent) . ',' . $item->id();
	}

	protected function map_children($item)
	{
		return $this->map_descendants($item)->where($this->option('tree_alias') . '.level = :level', $item->level + 1);
	}

	protected function map_siblings($item, $include_self = false)
	{
		$m = $this->map_join_tree()->where($this->option('tree_alias') . '.parent_id = :pid', $item->parent_id)
			->where($this->option('tree_alias') . '.level = :level', $item->level);
		if (!$include_self) {
			$m->where(self::table_from($this) . '.id <> :id', $item->id());
		}
		return $m;
	}

	protected function map_descendants_all($items)
	{
		$ids = array();
		foreach ($items as $item) {
			if (is_object($item)) {
				$ids[] = $item->id();
			} else if (is_array($item)) {
				$ids[] = $item['id'];
			} else {
				$ids[] = (int) $item;
			}
		}
		return $this->map_join_tree()->in($this->option('tree_alias') . '.parent_id', $ids)->map_sort_tree();
	}

	protected function map_descendants($item)
	{
		return $this->map_join_tree()->where($this->option('tree_alias') . '.parent_id = :id', $item->id())->map_sort_tree();
	}

	protected function map_trail($item, $include_self = true)
	{
		return $this->map_join_tree('parent_id')->where($this->option('tree_alias') . '.entity_id = :id', $item->id())->map_sort_tree();
	}

	protected function map_parent($item)
	{
		return $this->map_join_tree()->where(self::table_from($this) . '.id = :id', $item->parent_id);
	}

	protected function map_sort_tree($add = '')
	{
		$alias = $this->option('tree_alias');
		return $this->order_by("{$alias}.level, {$alias}.ord" . (!empty($add) ? ",$add" : ''));
	}

	//TODO: join_mapper method to DB.ORM
	protected function map_join_tree($field = 'entity_id', $alias = 'tt')
	{
		$this->option('tree_alias', $alias);
		return $this
			->join('inner', $this->tree_table() . " {$alias}", "{$alias}.{$field} = " . self::table_from($this) . '.id')
			->calculate(array("{$alias}.level level", "{$alias}.ord ord"));
	}

	public function tree_table()
	{
		return $this->option('tree_table') ? $this->option('tree_table') : self::table_from($this) . '_tree';
	}

	protected function map_join_tree_data($field = 'entity_id', $alias = 'tt')
	{
		$self_table = self::table_from($this);
		return $this->join_tree($field, $alias)->where("{$alias}.parent_id = {$self_table}.parent_id");
	}

	protected function map_full_tree()
	{
		return $this->map_join_tree_data();
	}

	public function tree_data($item)
	{
		$tree = $this->get_tree_mapper()->where('entity_id = :id', $item->id)->where('parent_id = :parent_id', $item->parent_id)
			->select_first();
		$item['level'] = $tree['level'];
		$item['ord'] = $tree['ord'];
		return $tree;
	}

	public function get_tree_mapper()
	{
		$table = $this->tree_table();
		if (!$this->session->can_cmap($table)) {
			$this->session->submapper($table, 'Tree.DB.MPT.JoinMapper');
		}
		return $this->session->$table->table($table)->spawn();
	}

}

class Tree_DB_MPT_Entity extends DB_ORM_Entity
{
	use Tree_DB_EntityTrait;
	use Tree_DB_MPT_EntityTrait;
}

trait Tree_DB_MPT_EntityTrait
{

	public function row_set_parent_id($value)
	{
		$this->set('old_parent_id', $this->get('parent_id'));
		$this->set('parent_id', $value);
		
		return $this;
	}

	public function set_parent_id($value)
	{
		return $this->row_set_parent_id($value);
	}

	public function get_trail($include_self = true)
	{
		$trail = $this->get_mapper()->trail($this, $include_self)->select();
		$trail = (array)$trail;
		if ($include_self) {
			$trail[] = $this;
		}
		return $trail;
	}

	public function get_level()
	{
		if (!isset($this->attrs['level'])) {
			$this->get_tree_data();
		}
		return is_null($this->attrs['level']) ? 0 : $this->attrs['level'];
	}

	public function get_tree_data()
	{
		$data = $this->get_mapper()->tree_data($this);
		return $data;
	}

	public function after_insert()
	{
		if ($parent = $this->get_parent()) {
			$trail = $parent->get_trail(true);
			$level = count($trail);
			foreach ($trail as $parent) {
				$this->get_tree_mapper()->make_entity(array('entity_id' => $this->id(), 'parent_id' => $parent->id(), 'level' => $level))->insert();
			}
		} else {
			$this->get_tree_mapper()->make_entity(array('entity_id' => $this->id(), 'parent_id' => 0, 'level' => 0))->insert();
		}
		$this->old_parent_id = $this->parent_id;
		return parent::after_insert();
	}

	public function after_find()
	{
		$this->get_mapper()->tree_data($this);
		return parent::after_find();
	}

	//TODO: refactor and optimize if possible
	public function after_update()
	{
		if ($this->old_parent_id != $this->parent_id) {
			$new_parent = $this->get_parent();
			$this->parent_id = $this->old_parent_id;
			$old_parent = $this->get_parent();
			$descendants = $this->get_descendants();
			$descendants[] = $this;
			foreach ($descendants as $d) {
				$ids[] = $d->id();
			}
			// Very bad case:
			// ----------
			if ($new_parent && in_array($new_parent->id(), $ids)) {
				$new_parent->parent_id = $old_parent ? $old_parent->id() : 0;
				$new_parent->update();
				$this->old_parent_id = $this->parent_id;
				$this->parent_id = $new_parent->id();
				return $this->after_update();
			}
			// ----------
			$dist = ($new_parent ? $new_parent->level : -1) - ($old_parent ? $old_parent->level : -1);
			if ($old_parent) { //delete old trail data
				foreach ($old_parent->get_trail(true) as $p) {
					$this->get_tree_mapper()->in('entity_id', $ids)->where('parent_id = :pid', $p->id())->delete_all();
				}
			} else {
				$this->get_tree_mapper()->in('entity_id', $ids)->where('parent_id = :pid', 0)->delete_all();
			}
			if ($new_parent) { //add data for new trail
				$trail = $new_parent->get_trail(true);
				foreach ($trail as $p) {
					foreach ($descendants as $d) {
						$this->get_tree_mapper()->make_entity(array('entity_id' => $d->id(), 'parent_id' => $p->id(), 'level' => $d->level))->insert();
					}
				}
				$this->parent_id = $new_parent->id();
			} else {
				$this->get_tree_mapper()->make_entity(array('entity_id' => $this->id(), 'parent_id' => 0, 'level' => 0))->insert();
				$k = array_search($this->id(), $ids);
				if ($k !== false) {
					unset($ids[$k]);
				}
			}
			//update level in descendants
			$this->get_tree_mapper()->in('entity_id', $ids)->update_all(array(), array('level' => "level + $dist"));
			$this->old_parent_id = $this->parent_id;
		}
		return parent::after_update();
	}

	public function get_tree_mapper()
	{
		return $this->get_mapper()->get_tree_mapper()->spawn();
	}

	public function after_delete()
	{
		foreach ($this->get_mapper()->descendants($this) as $des) {
			$des->delete();
			$this->get_tree_mapper()->where('entity_id = :eid', $des->id())->delete_all();
		}
		$this->get_tree_mapper()->where('entity_id = :eid', $this->id())->delete_all();
		return parent::after_delete();
	}

}

class Tree_DB_MPT_JoinMapper extends DB_ORM_SQLMapper
{

	protected function setup()
	{
		$this
			->columns('id', 'entity_id', 'parent_id', 'level', 'ord')
			->classname('Tree.DB.MPT.JoinEntity')
			->key('id');
		return parent::setup();
	}

}

class Tree_DB_MPT_JoinEntity extends DB_ORM_Entity
{

	protected function setup()
	{
		return $this->assign(array(
				'entity_id' => 0,
				'parent_id' => 0,
				'ord' => 0,
				'level' => 0
			)
		);
	}

}